#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the setup command."""
import builtins
import os

import yaml

import orion.core
import orion.core.cli


class _mock_input:
    """Mock `input` to return a serie of strings."""

    def __init__(self, values):
        """Set the serie of values"""
        self.values = values

    def __call__(self, *args):
        """Pop one value at each"""
        return self.values.pop(0)


def test_creation_when_not_existing(monkeypatch, tmp_path):
    """Test if a configuration file is created when it does not exist."""
    config_path = str(tmp_path) + "/tmp_config.yaml"
    monkeypatch.setattr(orion.core, "DEF_CONFIG_FILES_PATHS", [config_path])
    monkeypatch.setattr(builtins, "input", _mock_input(['type', 'name', 'host']))

    try:
        os.remove(config_path)
    except FileNotFoundError:
        pass

    orion.core.cli.main(["db", "setup"])

    assert os.path.exists(config_path)

    with open(config_path, 'r') as output:
        content = yaml.safe_load(output)

    assert content == {"database": {"type": "type", "name": "name", "host": "host"}}


def test_creation_when_exists(monkeypatch, tmp_path):
    """Test if the configuration file is overwritten when it exists."""
    config_path = str(tmp_path) + "/tmp_config.yaml"
    monkeypatch.setattr(orion.core, "DEF_CONFIG_FILES_PATHS", [config_path])
    monkeypatch.setattr(builtins, "input", _mock_input(['y', 'type', 'name', 'host']))

    dump = {"database": {"type": "allo2", "name": "allo2", "host": "allo2"}}

    with open(config_path, 'w') as output:
        yaml.dump(dump, output)

    orion.core.cli.main(["db", "setup"])

    with open(config_path, 'r') as output:
        content = yaml.safe_load(output)

    assert content != dump


def test_stop_creation_when_exists(monkeypatch, tmp_path):
    """Test if the configuration file is overwritten when it exists."""
    config_path = str(tmp_path) + "/tmp_config.yaml"
    monkeypatch.setattr(orion.core, "DEF_CONFIG_FILES_PATHS", [config_path])
    monkeypatch.setattr(builtins, "input", _mock_input(['n']))

    dump = {"database": {"type": "allo2", "name": "allo2", "host": "allo2"}}

    with open(config_path, 'w') as output:
        yaml.dump(dump, output)

    orion.core.cli.main(["db", "setup"])

    with open(config_path, 'r') as output:
        content = yaml.safe_load(output)

    assert content == dump


def test_defaults(monkeypatch, tmp_path):
    """Test if the default values are used when nothing user enters nothing."""
    config_path = str(tmp_path) + "/tmp_config.yaml"
    monkeypatch.setattr(orion.core, "DEF_CONFIG_FILES_PATHS", [config_path])
    monkeypatch.setattr(builtins, "input", _mock_input(['', '', '']))

    orion.core.cli.main(["db", "setup"])

    with open(config_path, 'r') as output:
        content = yaml.safe_load(output)

    assert content == {"database": {"type": "mongodb", "name": "test", "host": "localhost"}}



# -*- coding: utf-8 -*-
"""Perform functional tests for the REST endpoint `/experiments`"""
import copy
import datetime

from orion.core.worker.trial import Trial
from orion.storage.base import get_storage

current_id = 0

base_experiment = dict(
    name='experiment-name',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'test-user',
              'orion_version': 'x.y.z',
              'datetime': datetime.datetime(1, 1, 1),
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
)

base_trial = {
    'experiment': None,
    'status': 'new',
    'worker': None,
    'submit_time': datetime.datetime(1, 1, 1),
    'start_time': datetime.datetime(1, 1, 1, second=10),
    'end_time': datetime.datetime(1, 1, 2),
    'heartbeat': None,
    'results': [
        {'name': 'loss', 'type': 'objective', 'value': 0.05},
        {'name': 'a', 'type': 'statistic', 'value': 10},
        {'name': 'b', 'type': 'statistic', 'value': 5},
    ],
    'params': [
        {'name': 'x',
         'type': 'real',
         'value': 10.0},
    ],
}


class TestCollection:
    """Tests the server's response on experiments/"""

    def test_no_experiments(self, client):
        """Tests that the API returns a positive response when no experiments are present"""
        response = client.simulate_get('/experiments')

        assert response.json == []
        assert response.status == "200 OK"

    def test_send_name_and_versions(self, client):
        """Tests that the API returns all the experiments with their name and version"""
        expected = [
            {'name': 'a', 'version': 1},
            {'name': 'b', 'version': 1}
        ]

        _add_experiment(name='a', version=1, _id=1)
        _add_experiment(name='b', version=1, _id=2)

        response = client.simulate_get('/experiments')

        assert response.json == expected
        assert response.status == "200 OK"

    def test_latest_versions(self, client):
        """Tests that the API return the latest versions of each experiment"""
        expected = [
            {'name': 'a', 'version': 3},
            {'name': 'b', 'version': 1}
        ]

        _add_experiment(name='a', version=1, _id=1)
        _add_experiment(name='a', version=3, _id=2)
        _add_experiment(name='a', version=2, _id=3)
        _add_experiment(name='b', version=1, _id=4)

        response = client.simulate_get('/experiments')

        assert response.json == expected
        assert response.status == "200 OK"


class TestItem:
    """Tests the server's response to experiments/:name"""

    def test_non_existent_experiment(self, client):
        """
        Tests that a 404 response is returned when the experiment
        doesn't exist in the database
        """
        response = client.simulate_get('/experiments/a')

        assert response.status == "404 Not Found"
        assert response.json == {
            'title': 'Experiment not found',
            'description': 'Experiment "a" does not exist'
        }

        _add_experiment(name='a', version=1, _id=1)
        response = client.simulate_get('/experiments/b')

        assert response.status == "404 Not Found"
        assert response.json == {
            'title': 'Experiment not found',
            'description': 'Experiment "b" does not exist'
        }

    def test_experiment_specification(self, client):
        """Tests that the experiment returned is following the specification"""
        _add_experiment(name='a', version=1, _id=1)
        _add_trial(experiment=1, id_override="ae8", status='completed')

        response = client.simulate_get('/experiments/a')

        assert response.status == "200 OK"

        assert response.json['name'] == "a"
        assert response.json['version'] == 1
        assert response.json['status'] == "not done"
        assert response.json['trialsCompleted'] == 1
        assert response.json['startTime'] == "0001-01-01 00:00:00"  # TODO
        assert response.json['endTime'] == "0001-01-02 00:00:00"  # TODO
        assert len(response.json['user'])
        assert response.json['orionVersion'] == "x.y.z"

        _assert_config(response.json['config'])
        _assert_best_trial(response.json['bestTrial'])

    def test_default_is_latest_version(self, client):
        """Tests that the latest experiment is returned when no version parameter exists"""
        _add_experiment(name='a', version=1, _id=1)
        _add_experiment(name='a', version=3, _id=2)
        _add_experiment(name='a', version=2, _id=3)

        response = client.simulate_get('/experiments/a')

        assert response.status == "200 OK"
        assert response.json['version'] == 3

    def test_specific_version(self, client):
        """Tests that the specified version of an experiment is returned"""
        _add_experiment(name='a', version=1, _id=1)
        _add_experiment(name='a', version=2, _id=2)
        _add_experiment(name='a', version=3, _id=3)

        response = client.simulate_get('/experiments/a?version=2')

        assert response.status == "200 OK"
        assert response.json['version'] == 2

    def test_unknown_parameter(self, client):
        """
        Tests that if an unknown parameter is specified in
        the query string, an error is returned even if the experiment doesn't exist.
        """
        response = client.simulate_get('/experiments/a?unknown=true')

        assert response.status == "400 Bad Request"
        assert response.json == {
            'title': 'Invalid parameter',
            'description': 'Parameter "unknown" is not supported. Expected parameter "version".'
        }

        _add_experiment(name='a', version=1, _id=1)

        response = client.simulate_get('/experiments/a?unknown=true')

        assert response.status == "400 Bad Request"
        assert response.json == {
            'title': 'Invalid parameter',
            'description': 'Parameter "unknown" is not supported. Expected parameter "version".'
        }


def _add_experiment(**kwargs):
    """Adds experiment to the dummy orion instance"""
    base_experiment.update(copy.deepcopy(kwargs))
    get_storage().create_experiment(base_experiment)


def _add_trial(**kwargs):
    """Add trials to the dummy orion instance"""
    base_trial.update(copy.deepcopy(kwargs))
    get_storage().register_trial(Trial(**base_trial))


def _assert_config(config):
    """Asserts properties of the ``config`` dictionary"""
    assert config['poolSize'] == 1
    assert config['maxTrials'] == 10

    algorithm = config['algorithm']
    assert algorithm['name'] == 'random'
    assert algorithm['seed'] == 1

    space = config['space']
    assert len(space) == 1
    assert space['x'] == 'uniform(0, 200)'


def _assert_best_trial(best_trial):
    """Verifies properties of the best trial"""
    assert best_trial['id'] == 'ae8'
    assert best_trial['submitTime'] == '0001-01-01 00:00:00'
    assert best_trial['startTime'] == '0001-01-01 00:00:10'
    assert best_trial['endTime'] == '0001-01-02 00:00:00'

    parameters = best_trial['parameters']
    assert len(parameters) == 1
    assert parameters['x'] == 10.0

    assert best_trial['objective'] == 0.05

    statistics = best_trial['statistics']
    assert statistics['a'] == 10
    assert statistics['b'] == 5



# -*- coding: utf-8 -*-
"""
:mod:`orion.storage.track` -- Track Storage Protocol
====================================================

.. module:: base
   :platform: Unix
   :synopsis: Implement a storage protocol to allow Orion to use track as a storage method

"""

from collections import defaultdict
import copy
import datetime
import hashlib
import logging
import sys
import warnings

import orion.core
from orion.core.io.database import DuplicateKeyError
from orion.core.utils.flatten import flatten, unflatten
from orion.core.worker.trial import Trial as OrionTrial, validate_status
from orion.storage.base import BaseStorageProtocol, FailedUpdate, get_uid

log = logging.getLogger(__name__)


# TODO: Remove this when factory is reworked
class Track:    # noqa: F811
    """Forward declaration because of a weird factory bug where Track is not found"""

    def __init__(self, uri):
        assert False, 'This should not be called'


HAS_TRACK = False
REASON = None
try:
    from track.client import TrackClient
    from track.persistence.utils import parse_uri
    from track.serialization import to_json
    from track.structure import CustomStatus, Status as TrackStatus
    from track.structure import Project, Trial as TrackTrial, TrialGroup
    from track.persistence.local import ConcurrentWrite
    from track.utils import ItemNotFound

    HAS_TRACK = True
except ImportError:
    REASON = 'Track is not installed'

except SyntaxError:
    major, minor, patch, _, _ = sys.version_info

    if minor < 6:
        REASON = 'Python is too old'
        log.warning('Track does not support python < 3.6!')
    else:
        raise


if HAS_TRACK:
    _status = [
        CustomStatus('new', TrackStatus.CreatedGroup.value + 1),
        CustomStatus('reserved', TrackStatus.CreatedGroup.value + 2),
    ]

    _status_dict = {
        s.name: s for s in _status
    }
    _status_dict['completed'] = TrackStatus.Completed
    _status_dict['interrupted'] = TrackStatus.Interrupted
    _status_dict['broken'] = TrackStatus.Broken
    _status_dict['suspended'] = TrackStatus.Suspended


def get_track_status(val):
    """Convert orion status to track status"""
    return _status_dict.get(val)


def convert_track_status(status):
    """Convert track status to orion status"""
    return status.name.lower()


def remove_leading_slash(name):
    """Remove leading slash"""
    # if name[0] == '/':
    #     return name[1:]
    # return name
    return name


def add_leading_slash(name):
    """Add leading slash"""
    # if name[0] == '/':
    #     return name
    # return '/' + name
    return name


def to_epoch(date):
    """Convert datetime class into seconds since epochs"""
    return (date - datetime.datetime(1970, 1, 1)).total_seconds()


class TrialAdapter:
    """Mock Trial, see `~orion.core.worker.trial.Trial`

    Parameters
    ----------
    storage_trial
        Track trial object

    orion_trial
        Orion trial object

    objective: str
        objective key

    """

    def __init__(self, storage_trial, orion_trial=None, objective=None):
        self.storage = copy.deepcopy(storage_trial)
        self.memory = orion_trial
        self.session_group = None
        self.objective_key = objective
        self.objectives_values = None
        self._results = []

    def _repr_values(self, values, sep=','):
        """Represent with a string the given values."""
        return

    def __str__(self):
        """Represent partially with a string."""
        param_rep = ','.join(map(lambda value: "{0.name}:{0.value}".format(value), self._params))
        ret = "TrialAdapter(uid={3}, experiment={0}, status={1}, params={2})".format(
            repr(self.experiment[:10]), repr(self.status), param_rep, self.storage.uid)
        return ret

    __repr__ = __str__

    @property
    def experiment(self):
        """See `~orion.core.worker.trial.Trial`"""
        if self.memory is not None:
            return self.memory.experiment
        return self.storage.group_id

    @property
    def hearbeat(self):
        """See `~orion.core.worker.trial.Trial`"""
        return datetime.datetime.utcfromtimestamp(self.storage.metadata.get('heartbeat', 0))

    @property
    def id(self):
        """See `~orion.core.worker.trial.Trial`"""
        return self.storage.uid

    @property
    def params(self):
        """See `~orion.core.worker.trial.Trial`"""
        if self.memory is not None:
            return self.memory.params

        return unflatten({param.name: param.value for param in self._params})

    @property
    def _params(self):
        """See `~orion.core.worker.trial.Trial`"""
        if self.memory is not None:
            return self.memory._params

        types = self.storage.metadata['params_types']
        params = flatten(self.storage.parameters)

        return [
            OrionTrial.Param(name=add_leading_slash(name), value=params.get(name), type=vtype)
            for name, vtype in types.items()
        ]

    @property
    def status(self):
        """See `~orion.core.worker.trial.Trial`"""
        if self.memory is not None:
            return self.memory.status

        return convert_track_status(self.storage.status)

    @status.setter
    def status(self, value):
        """See `~orion.core.worker.trial.Trial`"""
        self.storage.status = get_track_status(value)

        if self.memory is not None:
            self.memory.status = value

    def to_dict(self):
        """See `~orion.core.worker.trial.Trial`"""
        trial = copy.deepcopy(self.storage.metadata)
        trial.update({
            'results': [r.to_dict() for r in self.results],
            'params': [p.to_dict() for p in self._params],
            '_id': self.storage.uid,
            'submit_time': self.submit_time,
            'experiment': self.experiment,
            'status': self.status
        })

        trial.pop('_update_count', 0)
        trial.pop('metric_types', 0)
        trial.pop('params_types')

        return trial

    @property
    def lie(self):
        """See `~orion.core.worker.trial.Trial`"""
        # we do not lie like Orion does
        return None

    @property
    def objective(self):
        """See `~orion.core.worker.trial.Trial`"""
        def result(val):
            return OrionTrial.Result(name=self.objective_key, value=val, type='objective')

        if self.objective_key is None:
            raise RuntimeError('no objective key was defined!')

        self.objectives_values = []

        data = self.storage.metrics.get(self.objective_key)
        if data is None:
            return None

        # objective was pushed without step data (already sorted)
        if isinstance(data, list):
            self.objectives_values = data
            return result(self.objectives_values[-1])

        # objective was pushed with step data
        elif isinstance(data, dict):
            for k, v in self.storage.metrics[self.objective_key].items():
                self.objectives_values.append((int(k), v))

            self.objectives_values.sort(key=lambda x: x[0])
            return result(self.objectives_values[-1][1])

        return None

    @property
    def results(self):
        """See `~orion.core.worker.trial.Trial`"""
        self._results = []

        for k, values in self.storage.metrics.items():
            result_type = 'statistic'
            if k == self.objective_key:
                result_type = 'objective'

            if isinstance(values, dict):
                items = list(values.items())
                items.sort(key=lambda v: v[0])

                val = items[-1][1]
                self._results.append(OrionTrial.Result(name=k, type=result_type, value=val))
            elif isinstance(values, list):
                self._results.append(OrionTrial.Result(name=k, type=result_type, value=values[-1]))

        return self._results

    @property
    def hash_params(self):
        """See `~orion.core.worker.trial.Trial`"""
        return OrionTrial.compute_trial_hash(self, ignore_fidelity=True)

    @results.setter
    def results(self, value):
        """See `~orion.core.worker.trial.Trial`"""
        self._results = value

    @property
    def gradient(self):
        """See `~orion.core.worker.trial.Trial`"""
        return None

    @property
    def submit_time(self):
        """See `~orion.core.worker.trial.Trial`"""
        return datetime.datetime.utcfromtimestamp(self.storage.metadata.get('submit_time'))

    @property
    def end_time(self):
        """See `~orion.core.worker.trial.Trial`"""
        return datetime.datetime.utcfromtimestamp(self.storage.metadata.get('end_time'))

    @end_time.setter
    def end_time(self, value):
        """See `~orion.core.worker.trial.Trial`"""
        self.storage.metadata['end_time'] = value

    @property
    def heartbeat(self):
        """Trial Heartbeat"""
        heartbeat = self.storage.metadata.get('heartbeat')
        if heartbeat:
            return datetime.datetime.utcfromtimestamp(heartbeat)
        return None

    @property
    def parents(self):
        """See `~orion.core.worker.trial.Trial`"""
        return self.storage.metadata.get('parent', [])

    @parents.setter
    def parents(self, other):
        """See `~orion.core.worker.trial.Trial`"""
        self.storage.metadata['parent'] = other


def experiment_uid(exp=None, name=None, version=None):
    """Return an experiment uid from its name and version for Track"""
    if name is None:
        name = exp.name

    if version is None:
        version = exp.version

    sha = hashlib.sha256()
    sha.update(name.encode('utf8'))
    sha.update(bytes([version]))
    return sha.hexdigest()


class Track(BaseStorageProtocol):   # noqa: F811
    """Implement a generic protocol to allow Orion to communicate using
    different storage backend

    Parameters
    ----------
    uri: str
        Track backend to use for storage; the format is as follow
         `protocol://[username:password@]host1[:port1][,...hostN[:portN]]][/[database][?options]]`

    """

    def __init__(self, uri):
        if not HAS_TRACK:
            # We ignored the import error above in case we did not need track
            # but now that we do we can rethrow it
            raise ImportError('Track is not installed!')

        self.uri = uri
        self.options = parse_uri(uri)['query']

        self.client = TrackClient(uri)
        self.backend = self.client.protocol
        self.project = None
        self.group = None
        self.objective = self.options.get('objective')
        self.lies = dict()
        assert self.objective is not None, 'An objective should be defined!'

    def _get_project(self, name):
        if self.project is None:
            self.project = self.backend.get_project(Project(name=name))

            if self.project is None:
                self.project = self.backend.new_project(Project(name=name))

        assert self.project, 'Project should have been found'

    def create_experiment(self, config):
        """Insert a new experiment inside the database"""
        self._get_project(config['name'])

        self.group = self.backend.new_trial_group(
            TrialGroup(
                name=experiment_uid(name=config['name'], version=config['version']),
                project_id=self.project.uid,
                metadata=to_json(config)
            )
        )

        if self.group is None:
            raise DuplicateKeyError('Experiment was already created')

        config['_id'] = self.group.uid
        return config

    def update_experiment(self, experiment=None, uid=None, where=None, **kwargs):
        """See :func:`~orion.storage.BaseStorageProtocol.update_experiment`"""
        uid = get_uid(experiment, uid)

        self.group = self.backend.fetch_and_update_group({
            '_uid': uid
        }, 'set_group_metadata', **kwargs)

        return self.group

    def fetch_experiments(self, query, selection=None):
        """Fetch all experiments that match the query"""
        new_query = {}
        for k, v in query.items():
            if k == 'name':
                new_query['metadata.name'] = v

            elif k.startswith('metadata'):
                new_query['metadata.{}'.format(k)] = v

            elif k == '_id':
                new_query['_uid'] = v

            else:
                new_query[k] = v

        groups = self.backend.fetch_groups(new_query)

        experiments = []
        for group in groups:
            version = group.metadata.get('version', 0)

            # metadata is experiment config
            exp = group.metadata
            exp.update({
                '_id': group.uid,
                'version': version,
                'name': group.project_id,
            })

            experiments.append(exp)

        return experiments

    def register_trial(self, trial):
        """Create a new trial to be executed"""
        stamp = datetime.datetime.utcnow()
        trial.submit_time = stamp

        metadata = dict()
        # pylint: disable=protected-access
        metadata['params_types'] = {remove_leading_slash(p.name): p.type for p in trial._params}
        metadata['submit_time'] = to_json(trial.submit_time)
        metadata['end_time'] = to_json(trial.end_time)
        metadata['worker'] = trial.worker
        metadata['metric_types'] = {remove_leading_slash(p.name): p.type for p in trial.results}
        metadata['metric_types'][self.objective] = 'objective'
        heartbeat = to_json(trial.heartbeat)
        if heartbeat is None:
            heartbeat = 0
        metadata['heartbeat'] = heartbeat

        metrics = defaultdict(list)
        for p in trial.results:
            metrics[p.name] = [p.value]

        if self.project is None:
            self._get_project(self.group.project_id)

        trial = self.backend.new_trial(TrackTrial(
            _hash=trial.hash_name,
            status=get_track_status(trial.status),
            project_id=self.project.uid,
            group_id=self.group.uid,
            parameters=trial.params,
            metadata=metadata,
            metrics=metrics
        ), auto_increment=False)

        if trial is None:
            raise DuplicateKeyError('Was not able to register Trial!')

        return TrialAdapter(trial, objective=self.objective)

    def register_lie(self, trial):
        """Register a *fake* trial created by the strategist.

        The main difference between fake trial and original ones is the addition of a fake objective
        result, and status being set to completed. The id of the fake trial is different than the id
        of the original trial, but the original id can be computed using the hashcode on parameters
        of the fake trial. See mod:`orion.core.worker.strategy` for more information and the
        Strategist object and generation of fake trials.

        Parameters
        ----------
        trial: `Trial` object
            Fake trial to register in the database

        """
        warnings.warn('Track does not persist lies!')

        if trial.id in self.lies:
            raise DuplicateKeyError('Lie already exists')

        self.lies[trial.id] = trial
        return trial

    def _fetch_trials(self, query, *args, **kwargs):
        """Fetch all the trials that match the query"""
        def sort_key(item):
            submit_time = item.submit_time
            if submit_time is None:
                return 0
            return submit_time

        query = to_json(query)

        new_query = {}
        for k, v in query.items():
            if k == 'experiment':
                new_query['group_id'] = v

            elif k == 'heartbeat':
                new_query['metadata.heartbeat'] = v

            elif k == '_id':
                new_query['uid'] = v

            elif k == 'end_time':
                new_query['metadata.end_time'] = v

            elif k == 'status' and isinstance(v, str):
                new_query['status'] = get_track_status(v)

            else:
                new_query[k] = v

        trials = [
            TrialAdapter(t, objective=self.objective) for t in self.backend.fetch_trials(new_query)
        ]
        trials.sort(key=sort_key)
        return trials

    def retrieve_result(self, trial, *args, **kwargs):
        """Fetch the result from a given medium (file, db, socket, etc..) for a given trial and
        insert it into the trial object
        """
        if isinstance(trial, TrialAdapter):
            trial = trial.storage

        refreshed_trial = self.backend.get_trial(trial)[0]
        new_trial = TrialAdapter(refreshed_trial, objective=self.objective)

        assert new_trial.objective is not None, 'Trial should have returned an objective value!'

        log.info("trial objective is (%s: %s)", self.objective, new_trial.objective.value)
        return new_trial

    def fetch_pending_trials(self, experiment):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_pending_trials`"""
        pending_status = ['new', 'suspended', 'interrupted']
        pending_status = [get_track_status(s) for s in pending_status]

        query = dict(
            group_id=experiment.id,
            status={'$in': pending_status}
        )

        return self._fetch_trials(query)

    def set_trial_status(self, trial, status, heartbeat=None):
        """Update the trial status and the heartbeat

        Raises
        ------
        FailedUpdate
            The exception is raised if the status of the trial object
            does not match the status in the database

        """
        validate_status(status)
        try:
            result_trial = self.backend.fetch_and_update_trial({
                'uid': trial.id,
                'status': get_track_status(trial.status)
            }, 'set_trial_status', status=get_track_status(status))

        except ItemNotFound as e:
            raise FailedUpdate() from e

        trial.status = status
        return result_trial

    def fetch_trials(self, experiment=None, uid=None):
        """See :func:`~orion.storage.BaseStorageProtocol.fetch_trials`"""
        uid = get_uid(experiment, uid)

        return self._fetch_trials(dict(group_id=uid))

    def get_trial(self, trial=None, uid=None):
        """See :func:`~orion.storage.BaseStorageProtocol.get_trials`"""
        uid = get_uid(trial, uid)

        _hash, _rev = 0, 0
        data = uid.split('_', maxsplit=1)

        if len(data) == 1:
            _hash = data[0]

        elif len(data) == 2:
            _hash, _rev = data

        trials = self.backend.get_trial(TrackTrial(_hash=_hash, revision=_rev))

        if trials is None:
            return None

        assert len(trials) == 1
        return TrialAdapter(trials[0], objective=self.objective)

    def reserve_trial(self, experiment):
        """Select a pending trial and reserve it for the worker"""
        query = dict(
            group_id=experiment.id,
            status={'$in': ['new', 'suspended', 'interrupted']}
        )

        try:
            trial = self.backend.fetch_and_update_trial(
                query,
                'set_trial_status',
                status=get_track_status('reserved'))

        except ItemNotFound:
            return None

        if trial is None:
            return None

        return TrialAdapter(trial, objective=self.objective)

    def update_trials(self, experiment=None, uid=None, where=None, **kwargs):
        """See :func:`~orion.storage.BaseStorageProtocol.update_trials`"""
        uid = get_uid(experiment, uid)

        if where is None:
            where = dict()

        where['group_id'] = uid

        count = 0
        for trial in self._fetch_trials(where):
            count += self.update_trial(trial, **kwargs)

        return count

    _ignore_updates_for = {'results', 'params', '_id'}

    def update_trial(self, trial=None, uid=None, **kwargs):
        """Update the fields of a given trials

        Parameters
        ----------
        trial: Trial
            Trial object to update

        where: Optional[dict]
            constraint trial must respect

        kwargs: dict
            a dictionary of fields to update

        Returns
        -------
        returns 1 if the underlying storage was updated else 0

        """
        # Get a TrialAdapter
        trial = self.get_trial(trial=trial, uid=uid)

        try:
            if isinstance(trial, TrialAdapter):
                trial = trial.storage

            for key, value in kwargs.items():
                if key == 'status':
                    self.backend.set_trial_status(trial, get_track_status(value))
                elif key in self._ignore_updates_for:
                    continue
                else:
                    pair = {key: to_json(value)}
                    self.backend.log_trial_metadata(trial, **pair)

            return 1
        except ConcurrentWrite:
            return 0

    def fetch_lost_trials(self, experiment):
        """Fetch all trials that have a heartbeat older than
        some given time delta (5 minutes by default)
        """
        heartbeat = orion.core.config.worker.heartbeat
        threshold = to_epoch(datetime.datetime.utcnow() - datetime.timedelta(seconds=heartbeat * 5))
        lte_comparison = {'$lte': threshold}
        query = {
            'experiment': experiment.id,
            'status': 'reserved',
            'heartbeat': lte_comparison
        }

        return self._fetch_trials(query)

    def push_trial_results(self, trial):
        """Push the trial's results to the database"""
        # Track already pushed the info no need to do it here
        pass

    def fetch_noncompleted_trials(self, experiment):
        """Fetch all non completed trials"""
        query = dict(
            group_id=experiment.id,
            status={'$ne': get_track_status('completed')}
        )
        return self.backend.fetch_trials(query)

    def fetch_trials_by_status(self, experiment, status):
        """Fetch all trials with the given status"""
        trials = self._fetch_trials(dict(status=status, group_id=experiment.id))
        return trials

    def count_completed_trials(self, experiment):
        """Count the number of completed trials"""
        return len(self._fetch_trials(dict(status='completed', group_id=experiment.id)))

    def count_broken_trials(self, experiment):
        """Count the number of broken trials"""
        return len(self._fetch_trials(dict(status='broken', group_id=experiment.id)))

    def update_heartbeat(self, trial):
        """Update trial's heartbeat"""
        self.backend.log_trial_metadata(trial.storage,
                                        heartbeat=to_epoch(datetime.datetime.utcnow()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db` -- Functions that define console scripts for database
==============================================================================

.. module:: cli
   :platform: Unix
   :synopsis: Commands to maintain database

"""



# -*- coding: utf-8 -*-
"""
:mod:`orion.serving.experiments_resource` -- Module responsible for the experiments/ REST endpoint
==================================================================================================

.. module:: experiments_resource
   :platform: Unix
   :synopsis: Serves all the requests made to experiments/ REST endpoint
"""
import json
from typing import Optional

from falcon import Request, Response

from orion.core.worker.experiment import Experiment
from orion.core.worker.trial import Trial
from orion.serving.parameters import retrieve_experiment, verify_query_parameters
from orion.serving.responses import build_experiment_response, build_experiments_response
from orion.storage.base import get_storage


class ExperimentsResource(object):
    """Handle requests for the experiments/ REST endpoint"""

    def __init__(self):
        self.storage = get_storage()

    def on_get(self, req: Request, resp: Response):
        """Handle the GET requests for experiments/"""
        experiments = self.storage.fetch_experiments({})
        leaf_experiments = _find_latest_versions(experiments)

        response = build_experiments_response(leaf_experiments)
        resp.body = json.dumps(response)

    def on_get_experiment(self, req: Request, resp: Response, name: str):
        """
        Handle GET requests for experiments/:name where `name` is
        the user-defined name of the experiment
        """
        verify_query_parameters(req.params, ['version'])
        version = req.get_param_as_int('version')
        experiment = retrieve_experiment(name, version)

        status = _retrieve_status(experiment)
        algorithm = _retrieve_algorithm(experiment)
        best_trial = _retrieve_best_trial(experiment)

        response = build_experiment_response(experiment, status, algorithm, best_trial)
        resp.body = json.dumps(response)


def _find_latest_versions(experiments):
    """Find the latest versions of the experiments"""
    leaf_experiments = {}
    for experiment in experiments:
        name = experiment['name']
        version = experiment['version']

        if name in leaf_experiments:
            leaf_experiments[name] = max(leaf_experiments[name], version)
        else:
            leaf_experiments[name] = version
    return leaf_experiments


def _retrieve_status(experiment: Experiment) -> str:
    """
    Determines the status of an experiment.

    Returns
    -------
        "done" if the experiment is complete, otherwise "not done".
    """
    return "done" if experiment.is_done else "not done"


def _retrieve_algorithm(experiment: Experiment) -> dict:
    """Populates the `algorithm` key with the configuration of the experiment's algorithm."""
    algorithm_name = list(experiment.algorithms.configuration.keys())[0]

    result = {'name': algorithm_name}
    result.update(experiment.algorithms.configuration[algorithm_name])
    return result


def _retrieve_best_trial(experiment: Experiment) -> Optional[Trial]:
    """Constructs the view of the best trial if there is one"""
    if not experiment.stats:
        return None

    return experiment.get_trial(uid=experiment.stats['best_trials_id'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.experiment_builder`."""
import copy
import datetime

import pytest

from orion.algo.base import BaseAlgorithm
from orion.algo.space import Space
from orion.core.evc.adapters import BaseAdapter
from orion.core.io.database.ephemeraldb import EphemeralDB
from orion.core.io.database.pickleddb import PickledDB
import orion.core.io.experiment_builder as experiment_builder
import orion.core.utils.backward as backward
from orion.core.utils.exceptions import BranchingEvent, NoConfigurationError, RaceCondition
from orion.core.utils.singleton import update_singletons
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy
from orion.testing import OrionState


def count_experiments():
    """Count experiments in storage"""
    return len(get_storage().fetch_experiments({}))


@pytest.fixture
def space():
    """Build a space definition"""
    return {'x': 'uniform(-50,50)'}


@pytest.fixture()
def python_api_config():
    """Create a configuration without the cli fluff."""
    new_config = dict(
        name='supernaekei',
        version=1,
        space={'x': 'uniform(0,10)'},
        metadata={'user': 'tsirif',
                  'orion_version': 'XYZ',
                  'VCS': {"type": "git",
                          "is_dirty": False,
                          "HEAD_sha": "test",
                          "active_branch": None,
                          "diff_sha": "diff"}},
        max_trials=1000,
        working_dir='',
        algorithms={
            'dumbalgo': {
                'done': False,
                'judgement': None,
                'scoring': 0,
                'seed': None,
                'suspend': False,
                'value': 5}},
        producer={'strategy': 'NoParallelStrategy'},
        _id='fasdfasfa',
        something_to_be_ignored='asdfa',
        refers=dict(
            root_id='supernaekei',
            parent_id=None,
            adapter=[])
        )

    return new_config


@pytest.fixture()
def new_config(random_dt, script_path):
    """Create a configuration that will not hit the database."""
    new_config = dict(
        name='supernaekei',
        metadata={'user': 'tsirif',
                  'orion_version': 'XYZ',
                  'user_script': script_path,
                  'user_config': 'abs_path/hereitis.yaml',
                  'user_args': [
                      script_path,
                      '--mini-batch~uniform(32, 256, discrete=True)'],
                  'VCS': {"type": "git",
                          "is_dirty": False,
                          "HEAD_sha": "test",
                          "active_branch": None,
                          "diff_sha": "diff"}},
        version=1,
        pool_size=10,
        max_trials=1000,
        working_dir='',
        algorithms={
            'dumbalgo': {
                'done': False,
                'judgement': None,
                'scoring': 0,
                'seed': None,
                'suspend': False,
                'value': 5}},
        producer={'strategy': 'NoParallelStrategy'},
        # attrs starting with '_' also
        _id='fasdfasfa',
        # and in general anything which is not in Experiment's slots
        something_to_be_ignored='asdfa',
        refers=dict(
            root_id='supernaekei',
            parent_id=None,
            adapter=[])
        )

    backward.populate_space(new_config)

    return new_config


@pytest.fixture
def parent_version_config():
    """Return a configuration for an experiment."""
    config = dict(
        _id='parent_config',
        name="old_experiment",
        version=1,
        algorithms='random',
        metadata={'user': 'corneauf', 'datetime': datetime.datetime.utcnow(),
                  'user_args': ['--x~normal(0,1)']})

    backward.populate_space(config)

    return config


@pytest.fixture
def child_version_config(parent_version_config):
    """Return a configuration for an experiment."""
    config = copy.deepcopy(parent_version_config)
    config['_id'] = 'child_config'
    config['version'] = 2
    config['refers'] = {'parent_id': 'parent_config'}
    config['metadata']['datetime'] = datetime.datetime.utcnow()
    config['metadata']['user_args'].append('--y~+normal(0,1)')
    backward.populate_space(config)
    return config


@pytest.mark.usefixtures('with_user_tsirif', 'version_XYZ')
def test_get_cmd_config(config_file):
    """Test local config (cmdconfig, cmdargs)"""
    cmdargs = {"config": config_file}
    local_config = experiment_builder.get_cmd_config(cmdargs)

    assert local_config['algorithms'] == 'random'
    assert local_config['strategy'] == 'NoParallelStrategy'
    assert local_config['max_trials'] == 100
    assert local_config['name'] == 'voila_voici'
    assert local_config['pool_size'] == 1
    assert local_config['storage'] == {
        'database': {
            'host': 'mongodb://user:pass@localhost',
            'name': 'orion_test',
            'type': 'mongodb'}}
    assert local_config['metadata'] == {
        'orion_version': 'XYZ',
        'user': 'tsirif'}


@pytest.mark.usefixtures('with_user_tsirif', 'version_XYZ')
def test_get_cmd_config_from_incomplete_config(incomplete_config_file):
    """Test local config with incomplete user configuration file
    (default, env_vars, cmdconfig, cmdargs)

    This is to ensure merge_configs update properly the subconfigs
    """
    cmdargs = {"config": incomplete_config_file}
    local_config = experiment_builder.get_cmd_config(cmdargs)

    assert 'algorithms' not in local_config
    assert 'max_trials' not in local_config
    assert 'pool_size' not in local_config
    assert 'name' not in local_config['storage']['database']
    assert local_config['storage']['database']['host'] == 'mongodb://user:pass@localhost'
    assert local_config['storage']['database']['type'] == 'incomplete'
    assert local_config['name'] == 'incomplete'
    assert local_config['metadata'] == {
        'orion_version': 'XYZ',
        'user': 'tsirif'}


def test_fetch_config_from_db_no_hit():
    """Verify that fetch_config_from_db returns an empty dict when the experiment is not in db"""
    with OrionState(experiments=[], trials=[]):
        db_config = experiment_builder.fetch_config_from_db(name='supernaekei')

    assert db_config == {}


@pytest.mark.usefixtures('with_user_tsirif')
def test_fetch_config_from_db_hit(new_config):
    """Verify db config when experiment is in db"""
    with OrionState(experiments=[new_config], trials=[]):
        db_config = experiment_builder.fetch_config_from_db(name='supernaekei')

    assert db_config['name'] == new_config['name']
    assert db_config['refers'] == new_config['refers']
    assert db_config['metadata'] == new_config['metadata']
    assert db_config['pool_size'] == new_config['pool_size']
    assert db_config['max_trials'] == new_config['max_trials']
    assert db_config['algorithms'] == new_config['algorithms']
    assert db_config['metadata'] == new_config['metadata']


@pytest.mark.usefixtures("with_user_tsirif")
def test_build_view_from_args_no_hit(config_file):
    """Try building experiment view when not in db"""
    cmdargs = {'name': 'supernaekei', 'config': config_file}

    with OrionState(experiments=[], trials=[]):
        with pytest.raises(NoConfigurationError) as exc_info:
            experiment_builder.build_view_from_args(cmdargs)
        assert "No experiment with given name 'supernaekei' and version '*'" in str(exc_info.value)


@pytest.mark.usefixtures("with_user_tsirif")
def test_build_view_from_args_hit(config_file, random_dt, new_config):
    """Try building experiment view when in db"""
    cmdargs = {'name': 'supernaekei', 'config': config_file}

    with OrionState(experiments=[new_config], trials=[]):
        exp_view = experiment_builder.build_view_from_args(cmdargs)

    assert exp_view._id == new_config['_id']
    assert exp_view.name == new_config['name']
    assert exp_view.configuration['refers'] == new_config['refers']
    assert exp_view.metadata == new_config['metadata']
    assert exp_view.pool_size == new_config['pool_size']
    assert exp_view.max_trials == new_config['max_trials']
    assert exp_view.algorithms.configuration == new_config['algorithms']


@pytest.mark.usefixtures("with_user_tsirif")
def test_build_view_from_args_hit_no_conf_file(config_file, random_dt, new_config):
    """Try building experiment view when in db, and local config file of user script does
    not exist
    """
    cmdargs = {'name': 'supernaekei', 'config': config_file}
    new_config['metadata']['user_args'] += ['--config', new_config['metadata']['user_config']]

    with OrionState(experiments=[new_config], trials=[]) as cfg:
        exp_view = experiment_builder.build_view_from_args(cmdargs)

    assert exp_view._id == new_config['_id']
    assert exp_view.name == new_config['name']
    assert exp_view.configuration['refers'] == new_config['refers']
    assert exp_view.metadata == new_config['metadata']
    assert exp_view.pool_size == new_config['pool_size']
    assert exp_view.max_trials == new_config['max_trials']
    assert exp_view.algorithms.configuration == new_config['algorithms']


@pytest.mark.usefixtures("with_user_dendi")
def test_build_from_args_no_hit(config_file, random_dt, script_path, new_config):
    """Try building experiment when not in db"""
    cmdargs = {'name': 'supernaekei', 'config': config_file,
               'user_args': [script_path,
                             'x~uniform(0,10)']}

    with OrionState(experiments=[], trials=[]):
        with pytest.raises(NoConfigurationError) as exc_info:
            experiment_builder.build_view_from_args(cmdargs)
        assert "No experiment with given name 'supernaekei' and version '*'" in str(exc_info.value)

        exp = experiment_builder.build_from_args(cmdargs)

    assert exp.name == cmdargs['name']
    assert exp.configuration['refers'] == {'adapter': [], 'parent_id': None, 'root_id': exp._id}
    assert exp.metadata['datetime'] == random_dt
    assert exp.metadata['user'] == 'dendi'
    assert exp.metadata['user_script'] == cmdargs['user_args'][0]
    assert exp.metadata['user_args'] == cmdargs['user_args']
    assert exp.pool_size == 1
    assert exp.max_trials == 100
    assert exp.algorithms.configuration == {'random': {'seed': None}}


@pytest.mark.usefixtures("version_XYZ", "with_user_tsirif", "mock_infer_versioning_metadata")
def test_build_from_args_hit(old_config_file, script_path, new_config):
    """Try building experiment when in db (no branch)"""
    cmdargs = {'name': 'supernaekei',
               'config': old_config_file,
               'user_args': [script_path,
                             '--mini-batch~uniform(32, 256, discrete=True)']}

    with OrionState(experiments=[new_config], trials=[]):
        # Test that experiment already exists
        experiment_builder.build_view_from_args(cmdargs)

        exp = experiment_builder.build_from_args(cmdargs)

    assert exp._id == new_config['_id']
    assert exp.name == new_config['name']
    assert exp.version == 1
    assert exp.configuration['refers'] == new_config['refers']
    assert exp.metadata == new_config['metadata']
    assert exp.max_trials == new_config['max_trials']
    assert exp.algorithms.configuration == new_config['algorithms']


@pytest.mark.usefixtures("with_user_bouthilx")
def test_build_from_args_force_user(new_config):
    """Try building experiment view when in db"""
    cmdargs = {'name': new_config['name']}
    cmdargs['user'] = 'tsirif'
    with OrionState(experiments=[new_config], trials=[]):
        # Test that experiment already exists
        exp_view = experiment_builder.build_from_args(cmdargs)
    assert exp_view.metadata['user'] == 'tsirif'


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_build_from_args_debug_mode(script_path):
    """Try building experiment in debug mode"""
    update_singletons()
    experiment_builder.build_from_args(
        {'name': 'whatever',
         'user_args': [script_path, '--mini-batch~uniform(32, 256)']})

    storage = get_storage()

    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)

    update_singletons()

    experiment_builder.build_from_args(
        {'name': 'whatever',
         'user_args': [script_path, '--mini-batch~uniform(32, 256)'],
         'debug': True})
    storage = get_storage()

    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, EphemeralDB)


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_build_view_from_args_debug_mode(script_path):
    """Try building experiment view in debug mode"""
    update_singletons()

    # Can't build view if none exist. It's fine we only want to test the storage creation.
    with pytest.raises(NoConfigurationError):
        experiment_builder.build_view_from_args({'name': 'whatever'})

    storage = get_storage()

    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, PickledDB)

    update_singletons()

    # Can't build view if none exist. It's fine we only want to test the storage creation.
    with pytest.raises(NoConfigurationError):
        experiment_builder.build_view_from_args({'name': 'whatever', 'debug': True})

    storage = get_storage()

    assert isinstance(storage, Legacy)
    assert isinstance(storage._db, EphemeralDB)


@pytest.mark.usefixtures("with_user_tsirif", "version_XYZ")
def test_build_no_hit(config_file, random_dt, script_path):
    """Try building experiment from config when not in db"""
    name = 'supernaekei'
    space = {'x': 'uniform(0, 10)'}
    max_trials = 100

    with OrionState(experiments=[], trials=[]):

        with pytest.raises(NoConfigurationError) as exc_info:
            experiment_builder.build_view(name)
        assert "No experiment with given name 'supernaekei' and version '*'" in str(exc_info.value)

        exp = experiment_builder.build(name, space=space, max_trials=max_trials)

    assert exp.name == name
    assert exp.configuration['refers'] == {'adapter': [], 'parent_id': None, 'root_id': exp._id}
    assert exp.metadata == {
        'datetime': random_dt,
        'user': 'tsirif',
        'orion_version': 'XYZ'}
    assert exp.configuration['space'] == space
    assert exp.max_trials == max_trials
    assert not exp.is_done
    assert exp.algorithms.configuration == {'random': {'seed': None}}


def test_build_no_commandline_config():
    """Try building experiment with no commandline configuration."""
    with OrionState(experiments=[], trials=[]):
        with pytest.raises(NoConfigurationError):
            experiment_builder.build('supernaekei')


@pytest.mark.usefixtures("with_user_dendi", "mock_infer_versioning_metadata", "version_XYZ")
def test_build_hit(python_api_config):
    """Try building experiment from config when in db (no branch)"""
    name = 'supernaekei'

    with OrionState(experiments=[python_api_config], trials=[]):

        # Test that experiment already exists (this should fail otherwise)
        experiment_builder.build_view(name=name)

        exp = experiment_builder.build(**python_api_config)

    assert exp._id == python_api_config['_id']
    assert exp.name == python_api_config['name']
    assert exp.configuration['refers'] == python_api_config['refers']
    python_api_config['metadata']['user'] = 'dendi'
    assert exp.metadata == python_api_config['metadata']
    assert exp.max_trials == python_api_config['max_trials']
    assert exp.algorithms.configuration == python_api_config['algorithms']


@pytest.mark.usefixtures("with_user_tsirif", "version_XYZ")
def test_build_without_config_hit(python_api_config):
    """Try building experiment without commandline config when in db (no branch)"""
    name = 'supernaekei'

    with OrionState(experiments=[python_api_config], trials=[]):

        # Test that experiment already exists (this should fail otherwise)
        experiment_builder.build_view(name=name)

        exp = experiment_builder.build(name=name)

    assert exp._id == python_api_config['_id']
    assert exp.name == python_api_config['name']
    assert exp.configuration['refers'] == python_api_config['refers']
    assert exp.metadata == python_api_config['metadata']
    assert exp.max_trials == python_api_config['max_trials']
    assert exp.algorithms.configuration == python_api_config['algorithms']


@pytest.mark.usefixtures("with_user_tsirif", "version_XYZ")
def test_build_from_args_without_cmd(old_config_file, script_path, new_config):
    """Try building experiment without commandline when in db (no branch)"""
    name = 'supernaekei'

    cmdargs = {'name': name,
               'config': old_config_file}

    with OrionState(experiments=[new_config], trials=[]):
        # Test that experiment already exists (this should fail otherwise)
        experiment_builder.build_view_from_args(cmdargs)

        exp = experiment_builder.build_from_args(cmdargs)

    assert exp._id == new_config['_id']
    assert exp.name == new_config['name']
    assert exp.configuration['refers'] == new_config['refers']
    assert exp.metadata == new_config['metadata']
    assert exp.max_trials == new_config['max_trials']
    assert exp.algorithms.configuration == new_config['algorithms']


@pytest.mark.usefixtures("with_user_tsirif")
class TestExperimentVersioning(object):
    """Create new Experiment with auto-versioning."""

    def test_new_experiment_wout_version(self, space):
        """Create a new and never-seen-before experiment without a version."""
        with OrionState():
            exp = experiment_builder.build(name="exp_wout_version", space=space)

        assert exp.version == 1

    def test_new_experiment_w_version(self, space):
        """Create a new and never-seen-before experiment with a version."""
        with OrionState():
            exp = experiment_builder.build(name="exp_wout_version", version=1, space=space)

        assert exp.version == 1

    def test_backward_compatibility_no_version(self, parent_version_config):
        """Branch from parent that has no version field."""
        parent_version_config.pop('version')
        with OrionState(experiments=[parent_version_config]):
            exp = experiment_builder.build(name=parent_version_config["name"],
                                           space={'y': 'uniform(0, 10)'})

        assert exp.version == 2

    def test_old_experiment_wout_version(self, parent_version_config):
        """Create an already existing experiment without a version."""
        with OrionState(experiments=[parent_version_config]):
            exp = experiment_builder.build(name=parent_version_config["name"])

        assert exp.version == 1

    def test_old_experiment_2_wout_version(self, parent_version_config, child_version_config):
        """Create an already existing experiment without a version and getting last one."""
        with OrionState(experiments=[parent_version_config, child_version_config]):
            exp = experiment_builder.build(name=parent_version_config["name"])

        assert exp.version == 2

    def test_old_experiment_w_version(self, parent_version_config, child_version_config):
        """Create an already existing experiment with a version."""
        with OrionState(experiments=[parent_version_config, child_version_config]):
            exp = experiment_builder.build(name=parent_version_config["name"], version=1)

        assert exp.version == 1

    def test_old_experiment_w_version_bigger_than_max(self, parent_version_config,
                                                      child_version_config):
        """Create an already existing experiment with a too large version."""
        print(child_version_config['name'])
        with OrionState(experiments=[parent_version_config, child_version_config]):
            exp = experiment_builder.build(name=parent_version_config["name"], version=8)

        assert exp.version == 2


@pytest.mark.usefixtures("with_user_tsirif", "version_XYZ")
class TestBuild(object):
    """Test building the experiment"""

    def test_good_set_before_init_hit_no_diffs_exc_max_trials(self, new_config):
        """Trying to set, and NO differences were found from the config pulled from db.

        Everything is normal, nothing changes. Experiment is resumed,
        perhaps with more trials to evaluate (an exception is 'max_trials').
        """
        with OrionState(experiments=[new_config], trials=[]):

            new_config['max_trials'] = 5000

            exp = experiment_builder.build(**new_config)

        # Deliver an external configuration to finalize init
        new_config['algorithms']['dumbalgo']['done'] = False
        new_config['algorithms']['dumbalgo']['judgement'] = None
        new_config['algorithms']['dumbalgo']['scoring'] = 0
        new_config['algorithms']['dumbalgo']['suspend'] = False
        new_config['algorithms']['dumbalgo']['value'] = 5
        new_config['algorithms']['dumbalgo']['seed'] = None
        new_config['producer']['strategy'] = "NoParallelStrategy"
        new_config.pop('something_to_be_ignored')
        assert exp.configuration == new_config

    def test_good_set_before_init_no_hit(self, random_dt, new_config):
        """Trying to set, overwrite everything from input."""
        with OrionState(experiments=[], trials=[]):
            exp = experiment_builder.build(**new_config)
            found_config = list(get_storage().fetch_experiments({'name': 'supernaekei',
                                                                 'metadata.user': 'tsirif'}))

        new_config['metadata']['datetime'] = exp.metadata['datetime']

        assert len(found_config) == 1
        _id = found_config[0].pop('_id')
        assert _id != 'fasdfasfa'
        assert exp._id == _id
        new_config['refers'] = {}
        new_config.pop('_id')
        new_config.pop('something_to_be_ignored')
        new_config['algorithms']['dumbalgo']['done'] = False
        new_config['algorithms']['dumbalgo']['judgement'] = None
        new_config['algorithms']['dumbalgo']['scoring'] = 0
        new_config['algorithms']['dumbalgo']['suspend'] = False
        new_config['algorithms']['dumbalgo']['value'] = 5
        new_config['algorithms']['dumbalgo']['seed'] = None
        new_config['refers'] = {'adapter': [], 'parent_id': None, 'root_id': _id}
        assert found_config[0] == new_config
        assert exp.name == new_config['name']
        assert exp.configuration['refers'] == new_config['refers']
        assert exp.metadata == new_config['metadata']
        assert exp.pool_size == new_config['pool_size']
        assert exp.max_trials == new_config['max_trials']
        assert exp.working_dir == new_config['working_dir']
        assert exp.version == new_config['version']
        assert exp.algorithms.configuration == new_config['algorithms']

    def test_working_dir_is_correctly_set(self, new_config):
        """Check if working_dir is correctly changed."""
        with OrionState():
            new_config['working_dir'] = './'
            exp = experiment_builder.build(**new_config)
            storage = get_storage()
            found_config = list(storage.fetch_experiments({'name': 'supernaekei',
                                                           'metadata.user': 'tsirif'}))

            found_config = found_config[0]
            exp = experiment_builder.build(**found_config)
            assert exp.working_dir == './'

    def test_working_dir_works_when_db_absent(self, database, new_config):
        """Check if working_dir is correctly when absent from the database."""
        with OrionState(experiments=[], trials=[]):
            exp = experiment_builder.build(**new_config)
            storage = get_storage()
            found_config = list(storage.fetch_experiments({'name': 'supernaekei',
                                                           'metadata.user': 'tsirif'}))

            found_config = found_config[0]
            exp = experiment_builder.build(**found_config)
            assert exp.working_dir == ''

    def test_configuration_hit_no_diffs(self, new_config):
        """Return a configuration dict according to an experiment object.

        Before initialization is done, it can be the case that the pair (`name`,
        user's name) has not hit the database. return a yaml compliant form
        of current state, to be used with :mod:`orion.core.cli.esolve_config`.
        """
        with OrionState(experiments=[new_config], trials=[]):
            experiment_count_before = count_experiments()
            exp = experiment_builder.build(**new_config)
            assert experiment_count_before == count_experiments()

        new_config['algorithms']['dumbalgo']['done'] = False
        new_config['algorithms']['dumbalgo']['judgement'] = None
        new_config['algorithms']['dumbalgo']['scoring'] = 0
        new_config['algorithms']['dumbalgo']['suspend'] = False
        new_config['algorithms']['dumbalgo']['value'] = 5
        new_config['algorithms']['dumbalgo']['seed'] = None
        new_config['producer']['strategy'] = "NoParallelStrategy"
        new_config.pop('something_to_be_ignored')
        assert exp.configuration == new_config

    def test_instantiation_after_init(self, new_config):
        """Verify that algo, space and refers was instanciated properly"""
        with OrionState(experiments=[new_config], trials=[]):
            exp = experiment_builder.build(**new_config)

        assert isinstance(exp.algorithms, BaseAlgorithm)
        assert isinstance(exp.space, Space)
        assert isinstance(exp.refers['adapter'], BaseAdapter)

    def test_algo_case_insensitive(self, new_config):
        """Verify that algo with uppercase or lowercase leads to same experiment"""
        with OrionState(experiments=[new_config], trials=[]):
            new_config['algorithms']['DUMBALGO'] = new_config['algorithms'].pop('dumbalgo')
            exp = experiment_builder.build(**new_config)

            assert exp.version == 1

    def test_hierarchical_space(self, new_config):
        """Verify space can have hierarchical structure"""
        space = {'a': {'x': 'uniform(0, 10, discrete=True)'},
                 'b': {'y': 'loguniform(1e-08, 1)',
                       'z': 'choices([\'voici\', \'voila\', 2])'}}

        with OrionState(experiments=[], trials=[]):
            exp = experiment_builder.build('hierarchy', space=space)

            exp2 = experiment_builder.build('hierarchy')

        assert 'a.x' in exp.space
        assert 'b.y' in exp.space
        assert 'b.z' in exp.space

        # Make sure it can be fetched properly from db as well
        assert 'a.x' in exp2.space
        assert 'b.y' in exp2.space
        assert 'b.z' in exp2.space

    def test_try_set_after_race_condition(self, new_config, monkeypatch):
        """Cannot set a configuration after init if it looses a race
        condition.

        The experiment from process which first writes to db is initialized
        properly. The experiment which looses the race condition cannot be
        initialized and needs to be rebuilt.
        """
        with OrionState(experiments=[new_config], trials=[]):
            experiment_count_before = count_experiments()

            def insert_race_condition(*args, **kwargs):
                if insert_race_condition.count == 0:
                    data = {}
                else:
                    data = new_config

                insert_race_condition.count += 1

                return data

            insert_race_condition.count = 0

            monkeypatch.setattr(experiment_builder, 'fetch_config_from_db', insert_race_condition)

            experiment_builder.build(**new_config)

            assert experiment_count_before == count_experiments()

        # Should be called
        # - once in build(),
        #     -> then register fails,
        # - then called once again in build,
        # - then called in build_view to evaluate the conflicts
        assert insert_race_condition.count == 3

    def test_algorithm_config_with_just_a_string(self):
        """Test that configuring an algorithm with just a string is OK."""
        name = 'supernaedo3'
        space = {'x': 'uniform(0,10)'}
        algorithms = 'dumbalgo'

        with OrionState(experiments=[], trials=[]):
            exp = experiment_builder.build(name=name, space=space, algorithms=algorithms)

        assert exp.configuration['algorithms'] == {
            'dumbalgo': {
                'done': False,
                'judgement': None,
                'scoring': 0,
                'suspend': False,
                'value': 5,
                'seed': None}}

    def test_new_child_with_branch(self):
        """Check that experiment is not incremented when branching with a new name."""
        name = 'parent'
        space = {'x': 'uniform(0, 10)'}

        with OrionState(experiments=[], trials=[]):
            parent = experiment_builder.build(name=name, space=space)

            assert parent.name == name
            assert parent.version == 1

            child_name = 'child'

            child = experiment_builder.build(name=name, branching={'branch_to': child_name})

            assert child.name == child_name
            assert child.version == 1
            assert child.refers['parent_id'] == parent.id

            child_name = 'child2'

            child = experiment_builder.build(name=child_name, branching={'branch_from': name})

            assert child.name == child_name
            assert child.version == 1
            assert child.refers['parent_id'] == parent.id

    def test_no_increment_when_child_exist(self):
        """Check that experiment cannot be incremented when asked for v1 while v2 exists."""
        name = 'parent'
        space = {'x': 'uniform(0,10)'}

        with OrionState(experiments=[], trials=[]):
            parent = experiment_builder.build(name=name, space=space)
            child = experiment_builder.build(name=name, space={'x': 'loguniform(1,10)'})
            assert child.name == parent.name
            assert parent.version == 1
            assert child.version == 2

            with pytest.raises(BranchingEvent) as exc_info:
                experiment_builder.build(name=name, version=1, space={'x': 'loguniform(1,10)'})
            assert 'Configuration is different and generates a branching' in str(exc_info.value)

    def test_race_condition_wout_version(self, monkeypatch):
        """Test that an experiment loosing the race condition during version increment raises
        RaceCondition if version number was not specified.
        """
        name = 'parent'
        space = {'x': 'uniform(0,10)'}

        with OrionState(experiments=[], trials=[]):
            parent = experiment_builder.build(name, space=space)
            child = experiment_builder.build(name=name, space={'x': 'loguniform(1,10)'})
            assert child.name == parent.name
            assert parent.version == 1
            assert child.version == 2

            # Either
            # 1.
            #     fetch_config_from_db only fetch parent
            #     test_version finds other child
            #     -> Detect race condition looking at conflicts
            # 2.
            #     fetch_config_from_db only fetch parent
            #     test_version do not find other child
            #     -> DuplicateKeyError

            def insert_race_condition_1(self, query):
                is_auto_version_query = (query == {'name': name, 'refers.parent_id': parent.id})
                if is_auto_version_query:
                    data = [child.configuration]
                # First time the query returns no other child
                elif insert_race_condition_1.count < 1:
                    data = [parent.configuration]
                else:
                    data = [parent.configuration, child.configuration]

                insert_race_condition_1.count += int(is_auto_version_query)

                return data

            insert_race_condition_1.count = 0

            monkeypatch.setattr(get_storage().__class__, 'fetch_experiments',
                                insert_race_condition_1)

            with pytest.raises(RaceCondition) as exc_info:
                experiment_builder.build(name=name, space={'x': 'loguniform(1,10)'})
            assert 'There was likely a race condition during version' in str(exc_info.value)

            def insert_race_condition_2(self, query):
                is_auto_version_query = (query == {'name': name, 'refers.parent_id': parent.id})
                # First time the query returns no other child
                if is_auto_version_query:
                    data = []
                elif insert_race_condition_2.count < 1:
                    data = [parent.configuration]
                else:
                    data = [parent.configuration, child.configuration]

                insert_race_condition_2.count += int(is_auto_version_query)

                return data

            insert_race_condition_2.count = 0

            monkeypatch.setattr(get_storage().__class__, 'fetch_experiments',
                                insert_race_condition_2)

            with pytest.raises(RaceCondition) as exc_info:
                experiment_builder.build(name=name, space={'x': 'loguniform(1,10)'})
            assert 'There was a race condition during branching.' in str(exc_info.value)

    def test_race_condition_w_version(self, monkeypatch):
        """Test that an experiment loosing the race condition during version increment cannot
        be resolved automatically if a version number was specified.

        Note that if we would raise RaceCondition, the conflict would still occur since
        the version number fetched will not be the new one from the resolution but the requested
        one. Therefore raising and handling RaceCondition would lead to infinite recursion in
        the experiment builder.
        """
        name = 'parent'
        space = {'x': 'uniform(0,10)'}

        with OrionState(experiments=[], trials=[]):
            parent = experiment_builder.build(name, space=space)
            child = experiment_builder.build(name=name, space={'x': 'loguniform(1,10)'})
            assert child.name == parent.name
            assert parent.version == 1
            assert child.version == 2

            # Either
            # 1.
            #     fetch_config_from_db only fetch parent
            #     test_version finds other child
            #     -> Detect race condition looking at conflicts
            # 2.
            #     fetch_config_from_db only fetch parent
            #     test_version do not find other child
            #     -> DuplicateKeyError

            def insert_race_condition_1(self, query):
                is_auto_version_query = (query == {'name': name, 'refers.parent_id': parent.id})
                if is_auto_version_query:
                    data = [child.configuration]
                # First time the query returns no other child
                elif insert_race_condition_1.count < 1:
                    data = [parent.configuration]
                else:
                    data = [parent.configuration, child.configuration]

                insert_race_condition_1.count += int(is_auto_version_query)

                return data

            insert_race_condition_1.count = 0

            monkeypatch.setattr(get_storage().__class__, 'fetch_experiments',
                                insert_race_condition_1)

            with pytest.raises(BranchingEvent) as exc_info:
                experiment_builder.build(name=name, version=1, space={'x': 'loguniform(1,10)'})
            assert 'Configuration is different and generates' in str(exc_info.value)

            def insert_race_condition_2(self, query):
                is_auto_version_query = (query == {'name': name, 'refers.parent_id': parent.id})
                # First time the query returns no other child
                if is_auto_version_query:
                    data = []
                elif insert_race_condition_2.count < 1:
                    data = [parent.configuration]
                else:
                    data = [parent.configuration, child.configuration]

                insert_race_condition_2.count += int(is_auto_version_query)

                return data

            insert_race_condition_2.count = 0

            monkeypatch.setattr(get_storage().__class__, 'fetch_experiments',
                                insert_race_condition_2)

            with pytest.raises(RaceCondition) as exc_info:
                experiment_builder.build(name=name, version=1, space={'x': 'loguniform(1,10)'})
            assert 'There was a race condition during branching.' in str(exc_info.value)


class TestInitExperimentView(object):
    """Create new ExperimentView instance."""

    def test_empty_experiment_view(self):
        """Hit user name, but exp_name does not hit the db."""
        with OrionState(experiments=[], trials=[]):
            with pytest.raises(NoConfigurationError) as exc_info:
                experiment_builder.build_view('supernaekei')
            assert ("No experiment with given name 'supernaekei' and version '*'"
                    in str(exc_info.value))

    def test_existing_experiment_view(self, new_config):
        """Hit exp_name + user's name in the db, fetch most recent entry."""
        with OrionState(experiments=[new_config], trials=[]):
            exp = experiment_builder.build_view(name='supernaekei')

        assert exp._id == new_config['_id']
        assert exp.name == new_config['name']
        assert exp.configuration['refers'] == new_config['refers']
        assert exp.metadata == new_config['metadata']
        assert exp.pool_size == new_config['pool_size']
        assert exp.max_trials == new_config['max_trials']
        assert exp.version == new_config['version']
        assert isinstance(exp.refers['adapter'], BaseAdapter)
        assert exp.algorithms.configuration == new_config['algorithms']

        with pytest.raises(AttributeError):
            exp.this_is_not_in_config = 5

        # Test that experiment.update_completed_trial indeed exists
        exp._experiment.update_completed_trial
        with pytest.raises(AttributeError):
            exp.update_completed_trial

        with pytest.raises(AttributeError):
            exp.register_trial

        with pytest.raises(AttributeError):
            exp.reserve_trial



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.strategies`."""
import logging

import pytest

from orion.core.worker.strategy import (
    MaxParallelStrategy, MeanParallelStrategy, NoParallelStrategy, Strategy, StubParallelStrategy)
from orion.core.worker.trial import Trial


@pytest.fixture
def observations():
    """10 objective observations"""
    points = [i for i in range(10)]
    results = [{'objective': points[i]} for i in range(10)]

    return points, results


@pytest.fixture
def incomplete_trial():
    """Return a single trial without results"""
    return Trial(params=[{'name': 'a', 'type': 'integer', 'value': 6}])


@pytest.fixture
def corrupted_trial():
    """Return a corrupted trial with results but status reserved"""
    return Trial(params=[{'name': 'a', 'type': 'integer', 'value': 6}],
                 results=[{'name': 'objective', 'type': 'objective', 'value': 1}],
                 status='reserved')


strategies = [
    'MaxParallelStrategy', 'MeanParallelStrategy', 'NoParallelStrategy', 'StubParallelStrategy']


@pytest.mark.parametrize('strategy', strategies)
def test_handle_corrupted_trials(caplog, strategy, corrupted_trial):
    """Verify that corrupted trials are handled properly"""
    with caplog.at_level(logging.WARNING, logger="orion.core.worker.strategy"):
        lie = Strategy(strategy).lie(corrupted_trial)

    match = "Trial `{}` has an objective but status is not completed".format(corrupted_trial.id)
    assert match in caplog.text

    assert lie is not None
    assert lie.value == corrupted_trial.objective.value


@pytest.mark.parametrize('strategy', strategies)
def test_handle_uncorrupted_trials(caplog, strategy, incomplete_trial):
    """Verify that no warning is logged if trial is valid"""
    with caplog.at_level(logging.WARNING, logger="orion.core.worker.strategy"):
        Strategy(strategy).lie(incomplete_trial)

    assert "Trial `{}` has an objective but status is not completed" not in caplog.text


class TestStrategyFactory:
    """Test creating a parallel strategy with the Strategy class"""

    def test_create_noparallel(self):
        """Test creating a NoParallelStrategy class"""
        strategy = Strategy('NoParallelStrategy')
        assert isinstance(strategy, NoParallelStrategy)

    def test_create_meanparallel(self):
        """Test creating a MeanParallelStrategy class"""
        strategy = Strategy('MeanParallelStrategy')
        assert isinstance(strategy, MeanParallelStrategy)


class TestParallelStrategies:
    """Test the different parallel strategy methods"""

    def test_max_parallel_strategy(self, observations, incomplete_trial):
        """Test that MaxParallelStrategy lies using the max"""
        points, results = observations

        strategy = MaxParallelStrategy()
        strategy.observe(points, results)
        lying_result = strategy.lie(incomplete_trial)

        max_value = max(result['objective'] for result in results)
        assert lying_result.value == max_value

    def test_mean_parallel_strategy(self, observations, incomplete_trial):
        """Test that MeanParallelStrategy lies using the mean"""
        points, results = observations

        strategy = MeanParallelStrategy()
        strategy.observe(points, results)
        lying_result = strategy.lie(incomplete_trial)

        mean_value = (sum(result['objective'] for result in results) /
                      float(len(results)))
        assert lying_result.value == mean_value

    def test_no_parallel_strategy(self, observations, incomplete_trial):
        """Test that NoParallelStrategy lies outputs None"""
        points, results = observations

        strategy = NoParallelStrategy()
        strategy.observe(points, results)
        lying_result = strategy.lie(incomplete_trial)
        assert lying_result is None

    def test_stub_parallel_strategy(self, observations, incomplete_trial):
        """Test that NoParallelStrategy lies outputs None"""
        points, results = observations

        strategy = StubParallelStrategy()
        strategy.observe(points, results)
        lying_result = strategy.lie(incomplete_trial)
        assert lying_result.value is None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.algo.base`."""

from orion.algo.base import BaseAlgorithm
from orion.algo.space import Integer, Real, Space


def test_init(dumbalgo):
    """Check if initialization works for nested algos."""
    nested_algo = {'DumbAlgo': dict(
        value=6,
        scoring=5
        )}
    algo = dumbalgo(8, value=1, subone=nested_algo)
    assert algo.space == 8
    assert algo.value == 1
    assert algo.scoring == 0
    assert algo.judgement is None
    assert algo.suspend is False
    assert algo.done is False
    assert isinstance(algo.subone, BaseAlgorithm)
    assert algo.subone.space == 8
    assert algo.subone.value == 6
    assert algo.subone.scoring == 5


def test_configuration(dumbalgo):
    """Check configuration getter works for nested algos."""
    nested_algo = {'DumbAlgo': dict(
        value=6,
        scoring=5
        )}
    algo = dumbalgo(8, value=1, subone=nested_algo)
    config = algo.configuration
    assert config == {
        'dumbalgo': {
            'seed': None,
            'value': 1,
            'scoring': 0,
            'judgement': None,
            'suspend': False,
            'done': False,
            'subone': {
                'dumbalgo': {
                    'seed': None,
                    'value': 6,
                    'scoring': 5,
                    'judgement': None,
                    'suspend': False,
                    'done': False,
                    }
                }
            }
        }


def test_space_setter(dumbalgo):
    """Check whether space setter works for nested algos."""
    nested_algo = {'DumbAlgo': dict(
        value=9,
        )}
    nested_algo2 = {'DumbAlgo': dict(
        judgement=10,
        )}
    algo = dumbalgo(8, value=1, naedw=nested_algo, naekei=nested_algo2)
    algo.space = 'etsh'
    assert algo.space == 'etsh'
    assert algo.naedw.space == 'etsh'
    assert algo.naedw.value == 9
    assert algo.naekei.space == 'etsh'
    assert algo.naekei.judgement == 10


def test_state_dict(dumbalgo):
    """Check whether trials_info is in the state dict"""
    nested_algo = {'DumbAlgo': dict(
        value=6,
        scoring=5
        )}
    algo = dumbalgo(8, value=1, subone=nested_algo)
    algo.suggest()
    assert not algo.state_dict['_trials_info']
    algo.observe([(1, 2)], [{'objective': 3}])
    assert len(algo.state_dict['_trials_info']) == 1
    algo.observe([(1, 2)], [{'objective': 3}])
    assert len(algo.state_dict['_trials_info']) == 1


def test_is_done_cardinality(monkeypatch, dumbalgo):
    """Check whether algorithm will stop with base algorithm cardinality check"""
    monkeypatch.delattr(dumbalgo, 'is_done')

    space = Space()
    space.register(Integer('yolo1', 'uniform', 1, 4))

    algo = dumbalgo(space)
    algo.suggest()
    for i in range(1, 6):
        algo.observe([[i]], [{'objective': 3}])

    assert len(algo.state_dict['_trials_info']) == 5
    assert algo.is_done

    space = Space()
    space.register(Real('yolo1', 'uniform', 1, 4))

    algo = dumbalgo(space)
    algo.suggest()
    for i in range(1, 6):
        algo.observe([[i]], [{'objective': 3}])

    assert len(algo.state_dict['_trials_info']) == 5
    assert not algo.is_done


def test_is_done_max_trials(monkeypatch, dumbalgo):
    """Check whether algorithm will stop with base algorithm max_trials check"""
    monkeypatch.delattr(dumbalgo, 'is_done')

    space = Space()
    space.register(Real('yolo1', 'uniform', 1, 4))

    algo = dumbalgo(space)
    algo.suggest()
    for i in range(1, 5):
        algo.observe([[i]], [{'objective': 3}])

    assert len(algo.state_dict['_trials_info']) == 4
    assert not algo.is_done

    dumbalgo.max_trials = 4
    assert algo.is_done



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker.producer` -- Produce and register samples to try
========================================================================

.. module:: producer
   :platform: Unix
   :synopsis: Suggest new parameter sets which optimize the objective.

"""
import copy
import logging
import random
import time

import orion.core
from orion.core.io.database import DuplicateKeyError
from orion.core.utils import format_trials
from orion.core.utils.exceptions import SampleTimeout, WaitingForTrials
from orion.core.worker.trial import Trial
from orion.core.worker.trials_history import TrialsHistory

log = logging.getLogger(__name__)


class Producer(object):
    """Produce suggested sets of problem's parameter space to try out.

    It uses an `Experiment` object to poll for not yet observed trials which
    have been already evaluated and to register new suggestions (points of
    the parameter `Space`) to be evaluated.

    """

    def __init__(self, experiment, max_idle_time=None):
        """Initialize a producer.

        :param experiment: Manager of this experiment, provides convenient
           interface for interacting with the database.
        """
        log.debug("Creating Producer object.")
        self.experiment = experiment
        self.space = experiment.space
        if self.space is None:
            raise RuntimeError("Experiment object provided to Producer has not yet completed"
                               " initialization.")
        self.algorithm = experiment.algorithms
        self.algorithm.algorithm.max_trials = experiment.max_trials
        if max_idle_time is None:
            max_idle_time = orion.core.config.worker.max_idle_time
        self.max_idle_time = max_idle_time
        self.strategy = experiment.producer['strategy']
        self.naive_algorithm = None
        # TODO: Move trials_history into PrimaryAlgo during the refactoring of Algorithm with
        #       Strategist and Scheduler.
        self.trials_history = TrialsHistory()
        self.params_hashes = set()
        self.naive_trials_history = None
        self.failure_count = 0

    @property
    def pool_size(self):
        """Pool-size of the experiment"""
        return self.experiment.pool_size

    def backoff(self):
        """Wait some time and update algorithm."""
        waiting_time = max(0, random.gauss(1, 0.2))
        log.info('Waiting %d seconds', waiting_time)
        time.sleep(waiting_time)
        log.info('Updating algorithm.')
        self.update()
        self.failure_count += 1

    def _sample_guard(self, start):
        """Check that the time taken sampling is less than max_idle_time"""
        if time.time() - start > self.max_idle_time:
            raise SampleTimeout(
                "Algorithm could not sample new points in less than {} seconds."
                "Failed to sample points {} times".format(self.max_idle_time, self.failure_count))

    def produce(self):
        """Create and register new trials."""
        sampled_points = 0

        # reset the number of time we failed to sample points
        self.failure_count = 0
        start = time.time()

        while (sampled_points < self.pool_size and
               not (self.experiment.is_done or self.naive_algorithm.is_done)):
            self._sample_guard(start)

            log.debug("### Algorithm suggests new points.")
            new_points = self.naive_algorithm.suggest(self.pool_size)

            # Sync state of original algo so that state continues evolving.
            self.algorithm.set_state(self.naive_algorithm.state_dict)

            if new_points is None:
                if self.algorithm.is_done:
                    return

                raise WaitingForTrials('Algo does not have more trials to sample.'
                                       'Waiting for current trials to finish')

            for new_point in new_points:
                sampled_points += self.register_trials(new_point)

    def register_trials(self, new_point):
        """Register a new set of sampled parameters into the DB
        guaranteeing their uniqueness

        Parameters
        ----------
        new_point: tuple
            tuple of values representing the hyperparameters values

        """
        # FIXME: Relying on DB to guarantee uniqueness
        # when the trial history will be held by that algo we can move that logic out of the DB

        log.debug("#### Convert point to `Trial` object.")
        new_trial = format_trials.tuple_to_trial(new_point, self.space)

        try:
            self._prevalidate_trial(new_trial)
            new_trial.parents = self.naive_trials_history.children
            log.debug("#### Register new trial to database: %s", new_trial)
            self.experiment.register_trial(new_trial)
            self._update_params_hashes([new_trial])
            return 1

        except DuplicateKeyError:
            log.debug("#### Duplicate sample.")
            self.backoff()
            return 0

    def _prevalidate_trial(self, new_trial):
        """Verify if trial is not in parent history"""
        if Trial.compute_trial_hash(new_trial, ignore_experiment=True) in self.params_hashes:
            raise DuplicateKeyError

    def _update_params_hashes(self, trials):
        """Register locally all param hashes of trials"""
        for trial in trials:
            self.params_hashes.add(
                Trial.compute_trial_hash(trial, ignore_experiment=True, ignore_lie=True))

    def update(self):
        """Pull all trials to update model with completed ones and naive model with non completed
        ones.
        """
        trials = self.experiment.fetch_trials(with_evc_tree=True)

        self._update_algorithm([trial for trial in trials if trial.status == 'completed'])
        self._update_naive_algorithm([trial for trial in trials if trial.status != 'completed'])

    def _update_algorithm(self, completed_trials):
        """Pull newest completed trials to update local model."""
        log.debug("### Fetch completed trials to observe:")

        new_completed_trials = []
        for trial in completed_trials:
            if trial not in self.trials_history:
                new_completed_trials.append(trial)

        log.debug("### %s", new_completed_trials)

        if new_completed_trials:
            log.debug("### Convert them to list of points and their results.")
            points = list(map(lambda trial: format_trials.trial_to_tuple(trial, self.space),
                              new_completed_trials))
            results = list(map(format_trials.get_trial_results, new_completed_trials))

            log.debug("### Observe them.")
            self.trials_history.update(new_completed_trials)
            self.algorithm.observe(points, results)
            self.strategy.observe(points, results)
            self._update_params_hashes(new_completed_trials)

    def _produce_lies(self, incomplete_trials):
        """Add fake objective results to incomplete trials

        Then register the trials in the db
        """
        log.debug("### Fetch active trials to observe:")
        lying_trials = []
        log.debug("### %s", incomplete_trials)

        for trial in incomplete_trials:
            log.debug("### Use defined ParallelStrategy to assign them fake results.")
            lying_result = self.strategy.lie(trial)
            if lying_result is not None:
                lying_trial = copy.deepcopy(trial)
                lying_trial.results.append(lying_result)
                lying_trials.append(lying_trial)
                log.debug("### Register lie to database: %s", lying_trial)
                lying_trial.parents = self.trials_history.children
                try:
                    self.experiment.register_lie(lying_trial)
                except DuplicateKeyError:
                    log.debug("#### Duplicate lie. No need to register a duplicate in DB.")

        return lying_trials

    def _update_naive_algorithm(self, incomplete_trials):
        """Pull all non completed trials to update naive model."""
        self.naive_algorithm = copy.deepcopy(self.algorithm)
        self.naive_trials_history = copy.deepcopy(self.trials_history)
        log.debug("### Create fake trials to observe:")
        lying_trials = self._produce_lies(incomplete_trials)
        log.debug("### %s", lying_trials)
        if lying_trials:
            log.debug("### Convert them to list of points and their results.")
            points = list(map(lambda trial: format_trials.trial_to_tuple(trial, self.space),
                              lying_trials))
            results = list(map(format_trials.get_trial_results, lying_trials))

            log.debug("### Observe them.")
            self.naive_trials_history.update(lying_trials)
            self.naive_algorithm.observe(points, results)
            self._update_params_hashes(lying_trials)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.serving.plots_resources` -- Module responsible for the plots/ REST endpoint
========================================================================================

.. module:: plots_resources
   :platform: Unix
   :synopsis: Serves all the requests made to plots/ REST endpoint
"""
from falcon import Request, Response

from orion.client import ExperimentClient
from orion.serving.parameters import retrieve_experiment
from orion.storage.base import get_storage


class PlotsResource(object):
    """Serves all the requests made to plots/ REST endpoint"""

    def __init__(self):
        self.storage = get_storage()

    def on_get_regret(self, req: Request, resp: Response, experiment_name: str):
        """
        Handle GET requests for plotting regret plots on plots/regret/:experiment
        where ``experiment`` is the user-defined name of the experiment.
        """
        experiment = ExperimentClient(retrieve_experiment(experiment_name), None)
        resp.body = experiment.plot.regret().to_json()



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.database` -- Wrappers for database frameworks
=================================================================

.. module:: database
   :platform: Unix
   :synopsis: Import name for wrappers of database frameworks.

Contains :class:`AbstractDB`, an interface for databases.
Currently, implemented wrappers:

   - :class:`orion.core.io.database.mongodb.MongoDB`

"""
from abc import abstractmethod, abstractproperty
import logging

from orion.core.utils.singleton import AbstractSingletonType, SingletonFactory


# pylint: disable=too-many-public-methods
class AbstractDB(object, metaclass=AbstractSingletonType):
    """Base class for database framework wrappers.

    Attributes
    ----------
    host : str
       It can be either:
          1. Known hostname or IP address in which database server resides.
          2. URI: A database framework specific connection string.
    name : str
       Name of database containing experiments.
    port : int
       Port that database server listens to for requests.
    username : str
        Name of user with write/read permissions to database with name `name`.
    password : str
        Secret phrase of user, `username`.

    """

    ASCENDING = 0
    DESCENDING = 1

    def __init__(self, host='localhost', name=None,
                 port=None, username=None, password=None, **kwargs):
        """Init method, see attributes of :class:`AbstractDB`."""
        self.host = host
        self.name = name
        self.port = port
        self.username = username
        self.password = password
        self.options = kwargs

        self._db = None
        self._conn = None
        self.initiate_connection()

    @abstractproperty
    def is_connected(self):
        """True, if practical connection has been achieved."""
        pass

    @abstractmethod
    def initiate_connection(self):
        """Connect to database, unless `AbstractDB` `is_connected`.

        Raises
        ------
        DatabaseError
            If connection or authentication fails

        """
        pass

    @abstractmethod
    def close_connection(self):
        """Disconnect from database, if `AbstractDB` `is_connected`."""
        pass

    @abstractmethod
    def ensure_index(self, collection_name, keys, unique=False):
        """Create given indexes if they do not already exist in database.

        Parameters
        ----------
        collection_name : str
           A collection inside database, a table.
        keys: str or list of tuples
           Can be a string representing a key to index, or a list of tuples
           with the structure `[(key_name, sort_order)]`. `key_name` must be a
           string and sort_order can be either `AbstractDB.ASCENDING` or
           AbstractDB.DESCENDING`.
        unique: bool, optional
           Ensure each document have a different key value. If not, operations
           like `write()` and `read_and_write()` will raise
           `DuplicateKeyError`.
           Defaults to False.

        Notes
        -----
        Depending on the backend, the indexing operation might operate in
        background. This means some operations on the database might occur
        before the indexes are totally built.

        """
        pass

    @abstractmethod
    def index_information(self, collection_name):
        """Return dict of names and sorting order of indexes

        Paramaters
        ----------
        collection_name : str
           A collection inside database, a table.

        Returns
        -------
        dict
            Dictionary of indexes where each key is the name in the format {name}_{order}
            and each value represents whether the index is unique.

        """
        return self._db[collection_name].index_information()

    @abstractmethod
    def drop_index(self, collection_name, name):
        """Remove index from the database

        Paramaters
        ----------
        collection_name : str
           A collection inside database, a table.
        name: str
            Index name in the format {name}_{order}

        """
        self._db[collection_name].drop_index(name)

    @abstractmethod
    def write(self, collection_name, data, query=None):
        """Write new information to a collection. Perform insert or update.

        Parameters
        ----------
        collection_name : str
           A collection inside database, a table.
        data : dict or list of dicts
           New data that will **be inserted** or that will **update** entries.
        query : dict, optional
           Assumes an update operation: filter entries in collection to be updated.

        Returns
        -------
        int
            Number of new documents if no query, otherwise number of modified documents.

        Notes
        -----
        In the case of an insert operation, `data` variable will be updated
        to contain a unique *_id* key.

        In the case of an update operation, if `query` fails to find a
        document that matches, no operation is performed.

        Raises
        ------
        DuplicateKeyError
            If the operation is creating duplicate keys in two different documents. Only occurs if
            the keys have unique indexes. See :meth:`AbstractDB.ensure_index` for more information
            about indexes.

        """
        pass

    @abstractmethod
    def read(self, collection_name, query=None, selection=None):
        """Read a collection and return a value according to the query.

        Parameters
        ----------
        collection_name : str
           A collection inside database, a table.
        query : dict, optional
           Filter entries in collection.
        selection : dict, optional
           Elements of matched entries to return, the projection.

        Returns
        -------
        list
            List of matched document[s]

        """
        pass

    @abstractmethod
    def read_and_write(self, collection_name, query, data, selection=None):
        """Read a collection's document and update the found document.

        If many documents are found, the first one is selected.

        Returns the updated document, or None if nothing found.

        Parameters
        ----------
        collection_name : str
           A collection inside database, a table.
        query : dict
           Filter entries in collection.
        data : dict or list of dicts
           New data that will **update** the entry.
        selection : dict, optional
           Elements of matched entries to return, the projection.

        Returns
        -------
        dict or None
            Updated first matched document or None if nothing found

        Raises
        ------
        DuplicateKeyError
            If the operation is creating duplicate keys in two different documents. Only occurs if
            the keys have unique indexes. See :meth:`AbstractDB.ensure_index` for more information
            about indexes.

        """
        pass

    @abstractmethod
    def count(self, collection_name, query=None):
        """Count the number of documents in a collection which match the `query`.

        Parameters
        ----------
        collection_name : str
           A collection inside database, a table.
        query : dict
           Filter entries in collection.

        """
        pass

    @abstractmethod
    def remove(self, collection_name, query):
        """Delete from a collection document[s] which match the `query`.

        Parameters
        ----------
        collection_name : str
           A collection inside database, a table.
        query : dict
           Filter entries in collection.

        Returns
        -------
        int
            Number of documents removed

        """
        pass


# pylint: disable=too-few-public-methods
class ReadOnlyDB(object):
    """Read-only view on a database."""

    __slots__ = ('_database', )

    #                     Attributes
    valid_attributes = (["host", "name", "port", "username", "password"] +
                        # Properties
                        ["is_connected"] +
                        # Methods
                        ["initiate_connection", "close_connection", "read", "count"])

    def __init__(self, database):
        """Init method, see attributes of :class:`AbstractDB`."""
        self._database = database

    def __getattr__(self, attr):
        """Get attribute only if valid"""
        if attr not in self.valid_attributes:
            raise AttributeError("Cannot access attribute %s on view-only experiments." % attr)

        return getattr(self._database, attr)


class DatabaseError(RuntimeError):
    """Exception type used to delegate responsibility from any database
    implementation's own Exception types.
    """

    pass


class DuplicateKeyError(DatabaseError):
    """Exception type used when a write attempt is made but the new document
    have an index already contained in the database.
    """

    pass


class DatabaseTimeout(DatabaseError):
    """Exception type used when there is a timeout during database operations."""

    pass


class OutdatedDatabaseError(DatabaseError):
    """Exception type used when the database is outdated."""

    pass


# pylint: disable=too-few-public-methods,abstract-method
class Database(AbstractDB, metaclass=SingletonFactory):
    """Class used to inject dependency on a database framework."""

    pass


# set per-module log level
logging.getLogger('filelock').setLevel('ERROR')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for :mod:`orion.algo.evolution_es`."""

import copy
import hashlib

import numpy as np
import pytest

from orion.algo.evolution_es import BracketEVES, compute_budgets, EvolutionES
from orion.algo.space import Fidelity, Real, Space


@pytest.fixture
def space():
    """Create a Space with a real dimension and a fidelity value."""
    space = Space()
    space.register(Real('lr', 'uniform', 0, 1))
    space.register(Fidelity('epoch', 1, 9, 1))
    return space


@pytest.fixture
def space1():
    """Create a Space with two real dimensions and a fidelity value."""
    space = Space()
    space.register(Real('lr', 'uniform', 0, 1))
    space.register(Real('weight_decay', 'uniform', 0, 1))
    space.register(Fidelity('epoch', 1, 8, 2))
    return space


@pytest.fixture
def space2():
    """Create a Space with two real dimensions."""
    space = Space()
    space.register(Real('lr', 'uniform', 0, 1))
    space.register(Real('weight_decay', 'uniform', 0, 1))
    return space


@pytest.fixture
def budgets():
    """Return a configuration for a bracket."""
    return [(30, 4), (30, 5), (30, 6)]


@pytest.fixture
def evolution(space1):
    """Return an instance of EvolutionES."""
    return EvolutionES(space1, repetitions=1, nums_population=4)


@pytest.fixture
def bracket(budgets, evolution, space1):
    """Return a `Bracket` instance configured with `b_config`."""
    return BracketEVES(evolution, budgets, 1, space1)


@pytest.fixture
def evolution_customer_mutate(space1):
    """Return an instance of EvolutionES."""
    return EvolutionES(space1, repetitions=1, nums_population=4,
                       mutate="orion.core.utils.tests.customized_mutate_example")


@pytest.fixture
def rung_0():
    """Create fake points and objectives for rung 0."""
    points = np.linspace(0, 8, 9)
    return dict(
        n_trials=9,
        resources=1,
        results={hashlib.md5(str([point]).encode('utf-8')).hexdigest(): (point, (1, point))
                 for point in points})


@pytest.fixture
def rung_1(rung_0):
    """Create fake points and objectives for rung 1."""
    values = map(lambda v: (v[0], (3, v[0])), list(sorted(rung_0['results'].values()))[:3])
    return dict(
        n_trials=3,
        resources=3,
        results={hashlib.md5(str([value[0]]).encode('utf-8')).hexdigest(): value
                 for value in values})


@pytest.fixture
def rung_2(rung_1):
    """Create fake points and objectives for rung 2."""
    values = map(lambda v: (v[0], (9, v[0])), list(sorted(rung_1['results'].values()))[:1])
    return dict(
        n_trials=1,
        resources=9,
        results={hashlib.md5(str([value[0]]).encode('utf-8')).hexdigest(): value
                 for value in values})


@pytest.fixture
def rung_3():
    """Create fake points and objectives for rung 3."""
    points = np.linspace(1, 4, 4)
    return dict(
        n_trials=4,
        resources=1,
        results={hashlib.md5(str([point]).encode('utf-8')).hexdigest():
                 (point, (np.power(2, (point - 1)), 1.0 / point, 1.0 / (point * point)))
                 for point in points})


@pytest.fixture
def rung_4():
    """Create duplicated fake points and objectives for rung 4."""
    points = np.linspace(1, 4, 4)
    return dict(
        n_trials=4,
        resources=1,
        results={hashlib.md5(str([point]).encode('utf-8')).hexdigest():
                 (point, (1, point // 2, point // 2))
                 for point in points})


def test_compute_budgets():
    """Verify proper computation of budgets on a logarithmic scale"""
    # Check typical values
    assert compute_budgets(1, 3, 1, 2, 1) == [[(2, 1), (2, 2), (2, 3)]]
    assert compute_budgets(1, 4, 2, 4, 2) == [[(4, 1), (4, 2), (4, 4)]]


def test_customized_mutate_population(space1, rung_3, budgets):
    """Verify customized mutated candidates is generated correctly."""
    customerized_dict = {'function': 'orion.testing.state.customized_mutate_example',
                         'multiply_factor': 2.0, 'add_factor': 1}
    algo = EvolutionES(space1, repetitions=1, nums_population=4,
                       mutate=customerized_dict)
    algo.brackets[0] = BracketEVES(algo, budgets, 1, space1)

    red_team = [0, 2]
    blue_team = [1, 3]
    population_range = 4
    for i in range(4):
        for j in [1, 2]:
            algo.brackets[0].eves.population[j][i] = list(rung_3["results"].values())[i][1][j]
        algo.brackets[0].eves.performance[i] = list(rung_3["results"].values())[i][0]

    org_data = np.stack((list(algo.brackets[0].eves.population.values())[0],
                         list(algo.brackets[0].eves.population.values())[1]), axis=0).T

    org_data = copy.deepcopy(org_data)

    algo.brackets[0]._mutate_population(red_team, blue_team,
                                        rung_3["results"], population_range)

    mutated_data = np.stack((list(algo.brackets[0].eves.population.values())[0],
                             list(algo.brackets[0].eves.population.values())[1]), axis=0).T

    # Winner team will be [0, 2], so [0, 2] will be remained, [1, 3] will be mutated.
    assert org_data.shape == mutated_data.shape
    assert (mutated_data[0] == org_data[0]).all()
    assert (mutated_data[2] == org_data[2]).all()
    assert (mutated_data[1] != org_data[1]).any()
    assert (mutated_data[3] != org_data[3]).any()
    assert (mutated_data[1] != org_data[0]).any()
    assert (mutated_data[3] != org_data[2]).any()

    # For each individual, mutation occurs in only one dimension chosen from two.
    # Customized test mutation function is divided by 2 for real type.
    if mutated_data[1][0] == org_data[0][0] / customerized_dict['multiply_factor']:
        assert mutated_data[1][1] == org_data[0][1]
    else:
        assert mutated_data[1][1] == org_data[0][1] / customerized_dict['multiply_factor']

    if mutated_data[3][0] == org_data[2][0] / customerized_dict['multiply_factor']:
        assert mutated_data[3][1] == org_data[2][1]
    else:
        assert mutated_data[3][1] == org_data[2][1] / customerized_dict['multiply_factor']


class TestEvolutionES():
    """Tests for the algo Evolution."""

    def test_register(self, evolution, bracket, rung_0, rung_1):
        """Check that a point is registered inside the bracket."""
        evolution.brackets = [bracket]
        bracket.hyperband = evolution
        bracket.eves = evolution
        bracket.rungs = [rung_0, rung_1]
        point = (1, 0.0)
        point_hash = hashlib.md5(str([0.0]).encode('utf-8')).hexdigest()

        evolution.observe([point], [{'objective': 0.0}])

        assert len(bracket.rungs[0])
        assert point_hash in bracket.rungs[0]['results']
        assert (0.0, point) == bracket.rungs[0]['results'][point_hash]


class TestBracketEVES():
    """Tests for `BracketEVES` class.."""

    def test_get_teams(self, bracket, rung_3):
        """Test that correct team is promoted."""
        bracket.rungs[0] = rung_3
        rung, population_range, red_team, blue_team = bracket._get_teams(0)
        assert len(list(rung.values())) == 4
        assert bracket.search_space_remove_fidelity == [1, 2]
        assert population_range == 4
        assert set(red_team).union(set(blue_team)) == {0, 1, 2, 3}
        assert set(red_team).intersection(set(blue_team)) == set()

    def test_mutate_population(self, bracket, rung_3):
        """Verify mutated candidates is generated correctly."""
        red_team = [0, 2]
        blue_team = [1, 3]
        population_range = 4
        for i in range(4):
            for j in [1, 2]:
                bracket.eves.population[j][i] = list(rung_3["results"].values())[i][1][j]
            bracket.eves.performance[i] = list(rung_3["results"].values())[i][0]

        org_data = np.stack((list(bracket.eves.population.values())[0],
                             list(bracket.eves.population.values())[1]), axis=0).T

        org_data = copy.deepcopy(org_data)

        bracket._mutate_population(red_team, blue_team,
                                   rung_3["results"], population_range)

        mutated_data = np.stack((list(bracket.eves.population.values())[0],
                                 list(bracket.eves.population.values())[1]), axis=0).T

        # Winner team will be [0, 2], so [0, 2] will be remained, [1, 3] will be mutated.
        assert org_data.shape == mutated_data.shape
        assert (mutated_data[0] == org_data[0]).all()
        assert (mutated_data[2] == org_data[2]).all()
        assert (mutated_data[1] != org_data[1]).any()
        assert (mutated_data[3] != org_data[3]).any()
        assert (mutated_data[1] != org_data[0]).any()
        assert (mutated_data[3] != org_data[2]).any()

        # For each individual, mutation occurs in only one dimension chosen from two.
        if mutated_data[1][0] != org_data[0][0]:
            assert mutated_data[1][1] == org_data[0][1]
        else:
            assert mutated_data[1][1] != org_data[0][1]

        if mutated_data[3][0] != org_data[2][0]:
            assert mutated_data[3][1] == org_data[2][1]
        else:
            assert mutated_data[3][1] != org_data[2][1]

    def test_duplicated_mutated_population(self, bracket, rung_4):
        """Verify duplicated candidates can be found and processed correctly."""
        red_team = [0, 2]
        blue_team = [0, 2]  # no mutate occur at first.
        population_range = 4
        for i in range(4):
            for j in [1, 2]:
                bracket.eves.population[j][i] = list(rung_4["results"].values())[i][1][j]
        points, nums_all_equal = bracket._mutate_population(red_team, blue_team,
                                                            rung_4["results"], population_range)

        # In this case, duplication will occur, and we can make it mutate one more time.
        # The points 1 and 2 should be different, while one of nums_all_equal should be 1.
        if points[1][1] != points[2][1]:
            assert points[1][2] == points[2][2]
        else:
            assert points[1][2] != points[2][2]

        assert nums_all_equal[0] == 0
        assert nums_all_equal[1] == 0
        assert nums_all_equal[2] == 1
        assert nums_all_equal[3] == 0

    def test_mutate_points(self, bracket, rung_3):
        """Test that correct point is promoted."""
        red_team = [0, 2]
        blue_team = [0, 2]
        population_range = 4
        for i in range(4):
            for j in [1, 2]:
                bracket.eves.population[j][i] = list(rung_3["results"].values())[i][1][j]
        points, nums_all_equal = bracket._mutate_population(red_team, blue_team,
                                                            rung_3["results"], population_range)
        assert points[0] == (1.0, 1.0, 1.0)
        assert points[1] == (2, 1.0 / 2, 1.0 / 4)
        assert (nums_all_equal == 0).all()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Common fixtures and utils for io tests."""


import copy
import os

import pytest

from orion.core.evc import conflicts


@pytest.fixture()
def config_file():
    """Open config file with new config"""
    file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             "orion_config.yaml")

    return open(file_path)


@pytest.fixture()
def old_config_file():
    """Open config file with original config from an experiment in db"""
    file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             "orion_old_config.yaml")

    return open(file_path)


@pytest.fixture()
def incomplete_config_file():
    """Open config file with partial database configuration"""
    file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             "orion_incomplete_config.yaml")

    return open(file_path)


@pytest.fixture
def parent_config():
    """Generate a new experiment configuration"""
    return dict(
        _id='test',
        name='test',
        metadata={'user': 'corneauf'},
        version=1)


@pytest.fixture
def child_config(parent_config):
    """Generate a new experiment configuration"""
    config = copy.deepcopy(parent_config)
    config['_id'] = "test2"
    config['refers'] = {'parent_id': 'test'}
    config['version'] = 2

    return config


@pytest.fixture
def experiment_name_conflict(create_db_instance, parent_config, child_config):
    """Generate an experiment name conflict"""
    create_db_instance.remove('experiments', {'name': 'test'})
    create_db_instance.remove('experiments', {'name': 'test2'})
    create_db_instance.write('experiments', parent_config)
    create_db_instance.write('experiments', child_config)
    return conflicts.ExperimentNameConflict(parent_config, parent_config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.checks.operations` -- Operations stage for database checks
===============================================================================

.. module:: operations
    :platform: Unix
    :synopsis: Checks for the operations of a `Database` object.

"""

from orion.core.utils.exceptions import CheckError


class OperationsStage:
    """The operations stage of the checks."""

    def __init__(self, creation_stage):
        """Create an intance of the stage.

        Parameters
        ----------
        creation_stage: `CreationStage`
            An instance of the previous stage.

        """
        self.c_stage = creation_stage

    def checks(self):
        """Return checks."""
        yield self.check_write
        yield self.check_read
        yield self.check_count
        yield self.check_remove

    def check_write(self):
        """Check if database supports write operation."""
        database = self.c_stage.instance

        try:
            database.write('test', {'index': 'value'})
        except Exception as ex:
            raise CheckError(str(ex))

        return "Success", ""

    def check_read(self):
        """Check if database supports read operation."""
        database = self.c_stage.instance

        try:
            result = database.read('test', {'index': 'value'})
        except Exception as ex:
            raise CheckError(str(ex))

        if not len(result):
            raise CheckError("Expected 'value', received nothing.")

        return "Success", ""

    def check_count(self):
        """Check if database supports count operation."""
        database = self.c_stage.instance

        count = database.count('test', {'index': 'value'})

        if count != 1:
            raise CheckError("Expected 1 hit, received {}.".format(count))

        return "Success", ""

    def check_remove(self):
        """Check if database supports delete operation."""
        database = self.c_stage.instance

        database.remove('test', {'index': 'value'})
        remaining = database.count('test', {'index': 'value'})

        if remaining:
            raise CheckError("{} items remaining.".format(remaining))

        return "Success", ""

    def post_stage(self):
        """Do nothing."""
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.io.database.pickleddb`."""
from datetime import datetime
import logging
from multiprocessing import Pool
import os

from filelock import FileLock, Timeout
import pytest

from orion.core.io.database import Database, DatabaseTimeout, DuplicateKeyError
from orion.core.io.database.ephemeraldb import EphemeralCollection
from orion.core.io.database.pickleddb import find_unpickable_doc, find_unpickable_field, PickledDB
import orion.core.utils.backward as backward


@pytest.fixture()
def orion_db():
    """Return PickledDB wrapper instance initiated with test opts."""
    PickledDB.instance = None
    orion_db = PickledDB(host='orion_db.pkl')
    yield orion_db


@pytest.fixture()
def clean_db(orion_db, exp_config):
    """Clean insert example experiment entries to collections."""
    if os.path.exists(orion_db.host):
        os.remove(orion_db.host)

    ephemeral_db = orion_db._get_database()
    database = ephemeral_db._db
    database['experiments'].drop()
    database['experiments'].insert_many(exp_config[0])
    database['trials'].drop()
    database['trials'].insert_many(exp_config[1])
    database['workers'].drop()
    database['workers'].insert_many(exp_config[2])
    database['resources'].drop()
    database['resources'].insert_many(exp_config[3])
    orion_db._dump_database(ephemeral_db)


@pytest.mark.usefixtures("clean_db")
class TestEnsureIndex(object):
    """Calls to :meth:`orion.core.io.database.pickleddb.PickledDB.ensure_index`."""

    def test_new_index(self, orion_db):
        """Index should be added to pickled database"""
        assert "new_field_1" not in orion_db._get_database()._db['new_collection']._indexes

        orion_db.ensure_index('new_collection', 'new_field', unique=False)
        assert "new_field_1" not in orion_db._get_database()._db['new_collection']._indexes

        orion_db.ensure_index('new_collection', 'new_field', unique=True)
        assert "new_field_1" in orion_db._get_database()._db['new_collection']._indexes

    def test_existing_index(self, orion_db):
        """Index should be added to pickled database and reattempt should do nothing"""
        assert "new_field_1" not in orion_db._get_database()._db['new_collection']._indexes

        orion_db.ensure_index('new_collection', 'new_field', unique=True)
        assert "new_field_1" in orion_db._get_database()._db['new_collection']._indexes

        # reattempt
        orion_db.ensure_index('new_collection', 'new_field', unique=True)
        assert "new_field_1" in orion_db._get_database()._db['new_collection']._indexes

    def test_compound_index(self, orion_db):
        """Tuple of Index should be added as a compound index."""
        assert "name_1_metadata.user_1" not in orion_db._get_database()._db['experiments']._indexes
        orion_db.ensure_index('experiments',
                              [('name', Database.ASCENDING),
                               ('metadata.user', Database.ASCENDING)], unique=True)
        assert "name_1_metadata.user_1" in orion_db._get_database()._db['experiments']._indexes


@pytest.mark.usefixtures("clean_db")
class TestRead(object):
    """Calls to :meth:`orion.core.io.database.pickleddb.PickledDB.read`."""

    def test_read_experiment(self, exp_config, orion_db):
        """Fetch a whole experiment entries."""
        loaded_config = orion_db.read(
            'trials', {'experiment': 'supernaedo2-dendi', 'status': 'new'})
        assert loaded_config == [exp_config[1][3], exp_config[1][4]]

        loaded_config = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': exp_config[1][3]['submit_time']})
        assert loaded_config == [exp_config[1][3]]
        assert loaded_config[0]['_id'] == exp_config[1][3]['_id']

    def test_read_with_id(self, exp_config, orion_db):
        """Query using ``_id`` key."""
        loaded_config = orion_db.read('experiments', {'_id': exp_config[0][2]['_id']})
        backward.populate_space(loaded_config[0])
        assert loaded_config == [exp_config[0][2]]

    def test_read_default(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'experiments', {'name': 'supernaedo2', 'metadata.user': 'tsirif'},
            selection={'algorithms': 1, '_id': 0})
        assert value == [{'algorithms': exp_config[0][0]['algorithms']}]

    def test_read_nothing(self, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'experiments', {'name': 'not_found', 'metadata.user': 'tsirif'},
            selection={'algorithms': 1})
        assert value == []

    def test_read_trials(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        value = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': {'$gte': datetime(2017, 11, 23, 0, 0, 0)}})
        assert value == [exp_config[1][1]] + exp_config[1][3:7]

        value = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'submit_time': {'$gt': datetime(2017, 11, 23, 0, 0, 0)}})
        assert value == exp_config[1][3:7]

    def test_null_comp(self, exp_config, orion_db):
        """Fetch value(s) from an entry."""
        all_values = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'end_time': {'$gte': datetime(2017, 11, 1, 0, 0, 0)}})

        db = orion_db._get_database()
        db._db['trials']._documents[0]._data['end_time'] = None
        orion_db._dump_database(db)

        values = orion_db.read(
            'trials',
            {'experiment': 'supernaedo2-dendi',
             'end_time': {'$gte': datetime(2017, 11, 1, 0, 0, 0)}})
        assert len(values) == len(all_values) - 1


@pytest.mark.usefixtures("clean_db")
class TestWrite(object):
    """Calls to :meth:`orion.core.io.database.pickleddb.PickledDB.write`."""

    def test_insert_one(self, orion_db):
        """Should insert a single new entry in the collection."""
        item = {'exp_name': 'supernaekei',
                'user': 'tsirif'}
        count_before = orion_db._get_database().count('experiments')
        # call interface
        assert orion_db.write('experiments', item) == 1
        assert orion_db._get_database().count('experiments') == count_before + 1
        value = orion_db._get_database()._db['experiments'].find({'exp_name': 'supernaekei'})[0]
        assert value == item

    def test_insert_many(self, orion_db):
        """Should insert two new entry (as a list) in the collection."""
        item = [{'exp_name': 'supernaekei2',
                 'user': 'tsirif'},
                {'exp_name': 'supernaekei3',
                 'user': 'tsirif'}]
        count_before = orion_db._get_database()._db['experiments'].count()
        # call interface
        assert orion_db.write('experiments', item) == 2
        database = orion_db._get_database()._db
        assert database['experiments'].count() == count_before + 2
        value = database['experiments'].find({'exp_name': 'supernaekei2'})[0]
        assert value == item[0]
        value = database['experiments'].find({'exp_name': 'supernaekei3'})[0]
        assert value == item[1]

    def test_update_many_default(self, orion_db):
        """Should match existing entries, and update some of their keys."""
        filt = {'metadata.user': 'dendi'}
        count_before = orion_db._get_database().count('experiments')
        count_query = orion_db._get_database().count('experiments', filt)
        # call interface
        assert orion_db.write('experiments', {'pool_size': 16}, filt) == count_query
        database = orion_db._get_database()._db
        assert database['experiments'].count() == count_before
        value = list(database['experiments'].find({}))
        assert value[0]['pool_size'] == 16
        assert value[1]['pool_size'] == 2
        assert value[2]['pool_size'] == 2
        assert value[3]['pool_size'] == 2

    def test_update_with_id(self, exp_config, orion_db):
        """Query using ``_id`` key."""
        filt = {'_id': exp_config[0][1]['_id']}
        count_before = orion_db._get_database().count('experiments')
        # call interface
        assert orion_db.write('experiments', {'pool_size': 36}, filt) == 1
        database = orion_db._get_database()._db
        assert database['experiments'].count() == count_before
        value = list(database['experiments'].find())
        assert value[0]['pool_size'] == 2
        assert value[1]['pool_size'] == 36
        assert value[2]['pool_size'] == 2


@pytest.mark.usefixtures("clean_db")
class TestReadAndWrite(object):
    """Calls to :meth:`orion.core.io.database.pickleddb.PickledDB.read_and_write`."""

    def test_read_and_write_one(self, orion_db, exp_config):
        """Should read and update a single entry in the collection."""
        # Make sure there is only one match
        documents = orion_db.read(
            'experiments',
            {'name': 'supernaedo4'})
        assert len(documents) == 1

        # Find and update atomically
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'name': 'supernaedo4'},
            {'pool_size': 'lalala'})
        exp_config[0][3]['pool_size'] = 'lalala'
        backward.populate_space(loaded_config)
        assert loaded_config == exp_config[0][3]

    def test_read_and_write_many(self, orion_db, exp_config):
        """Should update only one entry."""
        documents = orion_db.read('experiments', {'metadata.user': 'tsirif'})
        assert len(documents) > 1

        # Find many and update first one only
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'metadata.user': 'tsirif'},
            {'pool_size': 'lalala'})

        exp_config[0][1]['pool_size'] = 'lalala'
        backward.populate_space(loaded_config)
        assert loaded_config == exp_config[0][1]

        # Make sure it only changed the first document found
        documents = orion_db.read('experiments', {'metadata.user': 'tsirif'})
        assert documents[0]['pool_size'] == 'lalala'
        assert documents[1]['pool_size'] != 'lalala'

    def test_read_and_write_no_match(self, orion_db):
        """Should return None when there is no match."""
        loaded_config = orion_db.read_and_write(
            'experiments',
            {'name': 'lalala'},
            {'pool_size': 'lalala'})

        assert loaded_config is None

    def test_logging_when_getting_file_lock(self, caplog, orion_db):
        """When logging.level is ERROR, there should be no logging."""
        logging.basicConfig(level=logging.INFO)
        caplog.clear()
        caplog.set_level(logging.ERROR)
        # any operation will trigger the lock.
        orion_db.read(
            'experiments',
            {'name': 'supernaedo2', 'metadata.user': 'dendi'})

        assert 'acquired on orion_db.pkl.lock' not in caplog.text


@pytest.mark.usefixtures("clean_db")
class TestRemove(object):
    """Calls to :meth:`orion.core.io.database.pickleddb.PickledDB.remove`."""

    def test_remove_many_default(self, exp_config, orion_db):
        """Should match existing entries, and delete them all."""
        filt = {'metadata.user': 'tsirif'}
        database = orion_db._get_database()._db
        count_before = database['experiments'].count()
        count_filt = database['experiments'].count(filt)
        # call interface
        assert orion_db.remove('experiments', filt) == count_filt
        database = orion_db._get_database()._db
        assert database['experiments'].count() == count_before - count_filt
        assert database['experiments'].count() == 1
        loaded_config = list(database['experiments'].find())
        backward.populate_space(loaded_config[0])
        assert loaded_config == [exp_config[0][0]]

    def test_remove_with_id(self, exp_config, orion_db):
        """Query using ``_id`` key."""
        filt = {'_id': exp_config[0][0]['_id']}

        database = orion_db._get_database()._db
        count_before = database['experiments'].count()
        # call interface
        assert orion_db.remove('experiments', filt) == 1
        database = orion_db._get_database()._db
        assert database['experiments'].count() == count_before - 1
        loaded_configs = database['experiments'].find()
        for loaded_config in loaded_configs:
            backward.populate_space(loaded_config)
        assert loaded_configs == exp_config[0][1:]

    def test_remove_update_indexes(self, exp_config, orion_db):
        """Verify that indexes are properly update after deletion."""
        with pytest.raises(DuplicateKeyError):
            orion_db.write('experiments', {'_id': exp_config[0][0]['_id']})
        with pytest.raises(DuplicateKeyError):
            orion_db.write('experiments', {'_id': exp_config[0][1]['_id']})

        filt = {'_id': exp_config[0][0]['_id']}

        database = orion_db._get_database()._db
        count_before = database['experiments'].count()
        # call interface
        assert orion_db.remove('experiments', filt) == 1
        database = orion_db._get_database()._db
        assert database['experiments'].count() == count_before - 1
        # Should not fail now, otherwise it means the indexes were not updated properly during
        # remove()
        orion_db.write('experiments', filt)
        # And this should still fail
        with pytest.raises(DuplicateKeyError):
            orion_db.write('experiments', {'_id': exp_config[0][1]['_id']})


@pytest.mark.usefixtures("clean_db")
class TestCount(object):
    """Calls :meth:`orion.core.io.database.pickleddb.PickledDB.count`."""

    def test_count_default(self, exp_config, orion_db):
        """Call just with collection name."""
        found = orion_db.count('trials')
        assert found == len(exp_config[1])

    def test_count_query(self, exp_config, orion_db):
        """Call with a query."""
        found = orion_db.count('trials', {'status': 'completed'})
        assert found == len([x for x in exp_config[1] if x['status'] == 'completed'])

    def test_count_query_with_id(self, exp_config, orion_db):
        """Call querying with unique _id."""
        found = orion_db.count('trials', {'_id': exp_config[1][2]['_id']})
        assert found == 1

    def test_count_nothing(self, orion_db):
        """Call with argument that will not find anything."""
        found = orion_db.count('experiments', {'name': 'lalalanotfound'})
        assert found == 0


def write(field, i):
    """Write the given value to the pickled db."""
    PickledDB.instance = None
    orion_db = PickledDB(host='orion_db.pkl')
    try:
        orion_db.write('concurrent', {field: i})
    except DuplicateKeyError:
        print('dup')
        pass

    print(field, i)


@pytest.mark.usefixtures("clean_db")
class TestConcurreny(object):
    """Test concurrent operations"""

    def test_concurrent_writes(self, orion_db):
        """Test that concurrent writes all get written properly"""
        orion_db.ensure_index('concurrent', 'diff')

        assert orion_db.count('concurrent', {'diff': {'$gt': -1}}) == 0

        Pool(10).starmap(write, (('diff', i) for i in range(10)))

        assert orion_db.count('concurrent', {'diff': {'$gt': -1}}) == 10

    def test_concurrent_unique_writes(self, orion_db):
        """Test that concurrent writes cannot duplicate unique fields"""
        orion_db.ensure_index('concurrent', 'unique', unique=True)

        assert orion_db.count('concurrent', {'unique': 1}) == 0

        Pool(10).starmap(write, (('unique', 1) for i in range(10)))

        assert orion_db.count('concurrent', {'unique': 1}) == 1


def test_empty_file(orion_db):
    """Check that db loading can handle empty files"""
    with open(orion_db.host, 'wb') as f:
        f.write(b'')
    orion_db._get_database()


def test_unpickable_error_find_document():
    """Check error messages for pickledb"""
    class UnpickableClass:
        i_am_not_pickable = None

    unpickable_doc = {
        '_id': 2,
        'a_pickable': 1,
        'b_unpickable': UnpickableClass(),
        'c_pickable': 3
    }

    def make_pickable(uid):
        return {
            '_id': uid,
            'a_pickable': 1,
            'b_pickable': 2,
            'c_pickable': 3
        }

    unpickable_dict_of_dict = [
        make_pickable(1),
        unpickable_doc,
        make_pickable(3)
    ]

    pickable_dict_of_dict = [
        make_pickable(1),
        make_pickable(2),
        make_pickable(3)
    ]

    unpickable_collection = EphemeralCollection()
    unpickable_collection.insert_many(unpickable_dict_of_dict)

    pickable_collection = EphemeralCollection()
    pickable_collection.insert_many(pickable_dict_of_dict)

    database = {
        'pickable_collection': pickable_collection,
        'unpickable_collection': unpickable_collection
    }

    collection, doc = find_unpickable_doc(database)
    assert collection == 'unpickable_collection', 'should return the unpickable document'

    key, value = find_unpickable_field(doc)
    assert key == 'b_unpickable', 'should return the unpickable field'
    assert isinstance(value, UnpickableClass), 'should return the unpickable value'


def test_query_timeout(monkeypatch, orion_db):
    """Verify that filelock.Timeout is catched and reraised as DatabaseTimeout"""
    orion_db.timeout = 0.1

    def never_acquire(self, *arg, **kwargs):
        """Do not try to acquire, raise timeout"""
        raise Timeout(self)

    monkeypatch.setattr(FileLock, 'acquire', never_acquire)

    with pytest.raises(DatabaseTimeout) as exc:
        orion_db.read('whatever', {'it should': 'fail'})

    assert exc.match('Could not acquire lock for PickledDB after 0.1 seconds.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform a functional test of the hunt command."""
import pytest

import orion.core.cli


def test_hunt_no_prior(clean_db, one_experiment, capsys):
    """Test at least one prior is specified"""
    orion.core.cli.main(["hunt", "-n", "test", "./black_box.py"])

    captured = capsys.readouterr().err

    assert "No prior found" in captured
    assert 'Traceback' not in captured


def test_no_args(capsys):
    """Test that help is printed when no args are given."""
    with pytest.raises(SystemExit):
        orion.core.cli.main(['hunt'])

    captured = capsys.readouterr().out

    assert 'usage:' in captured
    assert 'Traceback' not in captured


def test_no_name(capsys):
    """Try to run the command without providing an experiment name"""
    returncode = orion.core.cli.main(["hunt", "--exp-max-trials", "10"])
    assert returncode == 1

    captured = capsys.readouterr().err

    assert captured == 'Error: No name provided for the experiment.\n'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests for db rm."""
import pytest

import orion.core.cli
from orion.storage.base import get_storage


def execute(command, assert_code=0):
    """Execute orion command and return returncode"""
    returncode = orion.core.cli.main(command.split(' '))
    assert returncode == assert_code


def test_no_exp(clean_db, capsys):
    """Test that rm non-existing exp exits gracefully"""
    execute('db rm i-dont-exist', assert_code=1)

    captured = capsys.readouterr()

    assert captured.err.startswith("Error: No experiment with given name 'i-dont-exist'")


def test_confirm_name(monkeypatch, clean_db, single_with_trials):
    """Test name must be confirmed for deletion"""
    def incorrect_name(*args):
        return 'oops'

    monkeypatch.setattr('builtins.input', incorrect_name)

    with pytest.raises(SystemExit):
        execute('db rm test_single_exp')

    def correct_name(*args):
        return 'test_single_exp'

    monkeypatch.setattr('builtins.input', correct_name)

    assert len(get_storage().fetch_experiments({})) == 1
    execute('db rm test_single_exp')
    assert len(get_storage().fetch_experiments({})) == 0


def test_one_exp(clean_db, single_with_trials):
    """Test that one exp is deleted properly"""
    assert len(get_storage().fetch_experiments({})) == 1
    assert len(get_storage()._fetch_trials({})) > 0
    execute('db rm -f test_single_exp')
    assert len(get_storage().fetch_experiments({})) == 0
    assert len(get_storage()._fetch_trials({})) == 0


def test_rm_all_evc(clean_db, three_family_branch_with_trials):
    """Test that deleting root removes all experiments"""
    assert len(get_storage().fetch_experiments({})) == 3
    assert len(get_storage()._fetch_trials({})) > 0
    execute('db rm -f test_double_exp --version 1')
    assert len(get_storage().fetch_experiments({})) == 0
    assert len(get_storage()._fetch_trials({})) == 0


def test_rm_under_evc(clean_db, three_family_branch_with_trials):
    """Test that deleting an experiment removes all children"""
    assert len(get_storage().fetch_experiments({})) == 3
    assert len(get_storage()._fetch_trials({})) > 0
    execute('db rm -f test_double_exp_child --version 1')
    assert len(get_storage().fetch_experiments({})) == 1
    assert len(get_storage()._fetch_trials({})) > 0
    # TODO: Test that the correct trials were deleted


def test_rm_default_leaf(clean_db, three_experiments_same_name):
    """Test that deleting an experiment removes the leaf by default"""
    assert len(get_storage().fetch_experiments({})) == 3
    execute('db rm -f test_single_exp')
    assert len(get_storage().fetch_experiments({})) == 2


def test_rm_trials_by_status(clean_db, single_with_trials):
    """Test that trials can be deleted by status"""
    trials = get_storage()._fetch_trials({})
    n_broken = sum(trial.status == 'broken' for trial in trials)
    assert n_broken > 0
    execute('db rm -f test_single_exp --status broken')
    assert len(get_storage()._fetch_trials({})) == len(trials) - n_broken


def test_rm_trials_all(clean_db, single_with_trials):
    """Test that trials all be deleted with '*'"""
    assert len(get_storage()._fetch_trials({})) > 0
    execute('db rm -f test_single_exp --status *')
    assert len(get_storage()._fetch_trials({})) == 0


def test_rm_trials_in_evc(clean_db, three_family_branch_with_trials):
    """Test that trials of parent experiment are not deleted"""
    assert len(get_storage().fetch_experiments({})) == 3
    assert len(get_storage()._fetch_trials({'experiment': 'test_double_exp_1'})) > 0
    assert len(get_storage()._fetch_trials({'experiment': 'test_double_exp_child_1'})) > 0
    assert len(get_storage()._fetch_trials({'experiment': 'test_double_exp_grand_child_1'})) > 0
    execute('db rm -f test_double_exp_child --status *')
    # Make sure no experiments were deleted
    assert len(get_storage().fetch_experiments({})) == 3
    # Make sure only trials of given experiment were deleted
    assert len(get_storage()._fetch_trials({'experiment': 'test_double_exp_1'})) > 0
    assert len(get_storage()._fetch_trials({'experiment': 'test_double_exp_child_1'})) == 0
    assert len(get_storage()._fetch_trials({'experiment': 'test_double_exp_grand_child_1'})) == 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.evc.cli`."""
import argparse

import pytest

from orion.core.cli.evc import get_branching_args_group
from orion.core.evc.conflicts import Resolution


def test_defined_parser():
    """Test that all expected branching arguments are present"""
    parser = argparse.ArgumentParser()
    get_branching_args_group(parser)

    options = parser.parse_args([])
    assert options.manual_resolution is False
    assert options.algorithm_change is False
    assert options.branch_to is None
    assert options.branch_from is None
    assert options.cli_change_type is None
    assert options.code_change_type is None
    assert options.config_change_type is None


def test_undefined_parser():
    """Test that creation of new resolution class make parser creation crash"""
    # It works
    get_branching_args_group(argparse.ArgumentParser())

    class DummyResolution(Resolution):
        ARGUMENT = "--dummy"

    # It doesn't
    with pytest.raises(AssertionError) as exc:
        get_branching_args_group(argparse.ArgumentParser())
    assert "A resolution with metavar 'dummy'" in str(exc.value)
